/*
  ddmap
  
  The concept of ddmap is the following: the iscsi-target keeps track of
  writes to facilitate the snapshot/read/replication process.

  Given 16KB segments if a write occurred in a segment, set a bit to 1. If you
  have a 1G disk, it would consist of 65536 segments. These segments can be encoded
  into a map of 8192 bytes.

  This userland ddmap tool allows you to obtain the header of the ddmap, the actual
  data of a ddmap, and it allows you to take the bits set to one and merge them
  with a map on disk.

  The map can be used by ddless to only read the actual changes from the snapshot
  instead of reading the entire snapshot source.
  
  The logic of getting a stable lvm snapshot is as follows:
  * dmsetup suspend device
  * create snapshot (background)
  * ddmap -s /proc/net/iet/ddmap/target.lun -t /vg_satabeast8_vol0/ddmap/target.lun
  * cp /vg_satabeast8_vol0/ddmap/target.lun /vg_satabeast8_vol0/ddmap/target.lun.snapshot
  * dmsetup resume device

  Now you can read data from the lvm snapshot and using the map you can restrict the
  reads to the regions that have changed.

  Steffen Plotner, 2010
*/

#include "ddless.h"
#include "dd_map.h"
#include "dd_log.h"
#include "dd_file.h"

//-----------------------------------------------------------------------------
// write the map
//-----------------------------------------------------------------------------
int ddmap_write(struct ddmap_data *source, struct ddmap_data *target)
{
	int ret = 0;
	int fd;
	
	struct ddmap_header *hdr = NULL;
	int header_bytes = sizeof(struct ddmap_header);
	int write_bytes;

	u_int32_t *source_map = source->map;
	u_int32_t *target_map = target->map;
	u_int32_t element;
	int i;

	if ((fd = open(target->map_device, O_CREAT|O_TRUNC|O_RDWR|O_LARGEFILE,
		S_IRUSR|S_IWUSR)) == -1 )
	{
		dd_log(LOG_ERR, "unable to open map: %s",target->map_device);
		return -1;
	}
	dd_log(LOG_INFO, "opened target map: %s",target->map_device);
	
	if ( (hdr = malloc(header_bytes)) == NULL )
	{
		dd_log(LOG_ERR,"unable to allocate memory for ddmap_header");
		return -1;
	}

	snprintf(hdr->info, sizeof(hdr->info), "*ddmap*");
	hdr->version = 1;
	hdr->name_sum = source->name_sum;
	hdr->map_size = source->map_size;
	
	if ( (write_bytes = write(fd, hdr, header_bytes)) == -1 )
	{
		dd_log(LOG_ERR, "unable to write header to map");
		ret = -1;
		goto err;
	}

	if ( write_bytes != header_bytes )
	{
		dd_log(LOG_ERR, "unable to write %d header bytes, got %d bytes",
			header_bytes, write_bytes);
		ret = -1;
		goto err;
	}

	dd_log(LOG_INFO,"source->map_size: %d target->map_size: %d", source->map_size, target->map_size);

	//
	// each loop processes a u32 worth of data (i.e. 4 bytes)
	//
	for (i=0; i<source->map_size; i++)
	{
		//
		// here is the core algorithm: OR between the source and the target
		//
		// we are using the target map if there is one and also if
		// the pointer from the source map is still within the range of the
		// current target map's pointer. This condition will generally always
		// meet EXCEPT when the either there is NO target map OR the source
		// map has grown.
		//
		if ( target->map && i < target->map_size )
		{
			dd_log(LOG_DEBUG,"i: %d source: %p target: %p", 
				i, source_map - source->map, target_map - target->map);
			element = *source_map | *target_map;
			source_map++;
			target_map++;
		}
		else
		{
			//
			// if there is no target (or source>target), then just read from the source
			//
			dd_log(LOG_DEBUG,"i: %d source: %p", i, source_map - source->map);
			element = *source_map;
			source_map++;
		}

		if ( (write_bytes = write(fd, &element, DDMAP_U32_SIZE)) == -1 )
		{
			dd_log(LOG_ERR, "unable to write header to map");
			ret = -1;
			goto err;
		}
		
		if ( write_bytes != DDMAP_U32_SIZE )
		{
			dd_log(LOG_ERR, "unable to write %d header bytes, got %d bytes",
				DDMAP_U32_SIZE, write_bytes);
			ret = -1;
			goto err;
		}
	}
	dd_log(LOG_INFO, "merge completed.");
err:
	if (hdr)
		free(hdr);
	if (fd)
		close(fd);
	return ret;
}

//-----------------------------------------------------------------------------
// help
//-----------------------------------------------------------------------------
void usage()
{
	printf(
"ddmap by Steffen Plotner, release date: "RELEASE_DATE"\n"
"\n"
"Reads ddmap provided by iet and dumps either the header, the data or merges\n"
"it with another file for permanent storage of the change map of the underlying\n"
"volume.\n"
"\n"
"	ddmap	-s <source> [-a] [-b] -t <target>\n"
"\n"
"Parameters\n"
"\n"
"	-s	ddmap source device\n"
"	-t	ddmap target device\n"
"	-a	dump the header\n"
"	-b	dump the data (hexdump)\n"
"	-v	verbose\n"
"	-vv	verbose+debug\n"
"\n"
"Exit codes:\n"
"	0	successful\n"
"	1	error\n"
);
}

//-----------------------------------------------------------------------------
// main
//-----------------------------------------------------------------------------
int main(int argc, char *argv[])
{
	int c, errflg;
	extern char *optarg;
	extern int optind, optopt;

	struct ddmap_data *map_source = NULL;
	struct ddmap_data *map_target = NULL;
	int ret = 0;
	int dump_header = 0;
	int dump_data = 0;
	
	dd_log_init("ddmap");
	
	if ( ( map_source = malloc(sizeof(struct ddmap_data))) == NULL )
	{
		dd_log(LOG_ERR, "unable allocated memory for map");
		ret = 1;
		goto err;
	}
	memset(map_source, 0, sizeof(struct ddmap_data));
	
	if ( ( map_target = malloc(sizeof(struct ddmap_data))) == NULL )
	{
		dd_log(LOG_ERR, "unable allocated memory for map");
		ret = 1;
		goto err;
	}
	memset(map_target, 0, sizeof(struct ddmap_data));

	errflg = 0;
	while ((c = getopt(argc, argv, "abs:t:hv")) != -1)
	{
		switch (c)
		{
			case 's':
				strncpy(map_source->map_device, optarg, DEV_NAME_LENGTH);
				break;
			case 't':
				strncpy(map_target->map_device, optarg, DEV_NAME_LENGTH);
				break;
			case 'a':
				dump_header = 1;
				break;
			case 'b':
				dump_data = 1;
				break;
			case 'v':
				dd_loglevel_inc();
				break;
			case 'h':
			case '?':
				errflg++;
		}
	}
	if (errflg)
	{
		usage();
		exit(1);
	}

	if ( *map_source->map_device && !*map_target->map_device )
	{
		if ( ddmap_read(map_source,dump_header) )
		{
			ret = -1;
			goto err;
		}
		if ( dump_data )
			ddmap_dump(map_source);
	}
	else if ( *map_source->map_device && *map_target->map_device )
	{
		//
		// read the source map
		//
		if ( ddmap_read(map_source,dump_header) )
		{
			ret = -1;
			goto err;
		}
		
		//
		// if there is no target map, forcably create it
		//
		if ( !dd_file_exists(map_target->map_device) )
		{
			dd_log(LOG_INFO, "target does not exist, creating new file");
			ddmap_write(map_source, map_target);
		}
		else
		{
			//
			// found target, then we have to read and verify it
			//
			if ( ddmap_read(map_target,dump_header) )
			{
				ret = -1;
				goto err;
			}

			//
			// verify the name sums - must match (data origin check)
			//
			if ( map_source->name_sum == map_target->name_sum )
			{
				dd_log(LOG_INFO, "name sum matches on source and target, good");
			}
			else
			{
				dd_log(LOG_ERR, "name sum mismatch: source 0x%08x target 0x%08x",
					map_source->name_sum, map_target->name_sum);
				ret = 1;
				goto err;
			}
			
			//
			// source size must be bigger than destination (device growth allowed)
			//
			if ( map_source->map_size >= map_target->map_size )
			{
				dd_log(LOG_INFO, "source is greater or equal to target size, good");
			}
			else
			{
				dd_log(LOG_ERR, "source size has shrunk?");
				ret = 1;
				goto err;
			}
			ddmap_write(map_source, map_target);
		}
	}
	else
	{
		usage();
		exit(1);
	}
err:	
	if (map_source->map)
		free(map_source->map);
	if (map_target->map)
		free(map_target->map);

	if (map_source)
		free(map_source);
	if (map_target)
		free(map_target);
	if ( ret == 0 )
		exit(0);
	exit(1);
}
