#!/usr/bin/perl

# Equallogic luncopy to ZFS, by Steffen Plotner, 2005, 2006, 2008, 2009

use strict;
use Data::Dumper;
use Getopt::Std;
use File::Copy;
use Sys::Hostname;

use FindBin;
use lib "$FindBin::Bin/lib";
use lib "$FindBin::Bin/etc";

# vendor provided
use EqlScript;

# amherst provided
use common;
use config;
use eqlutil;
use SunIscsiInitiator;
use SunDiskInfo;
use SunZfs;
use Sun2Tb;

$cfg{'verbose'} = 1;

#
# handle command line options
#
my %opts;
getopts('hg:s:f:',\%opts);
if ( defined($opts{'h'}) )
{
	print <<EOH;
$0 [-h] [-g eqlogicname] -s volume | -f ddless_filter,...

-h	help
-g	equallogic name (must be defined in eql.conf, default $cfg{'default_eqlogic'})
-s	process a specific volume
-f	filter on DDLESS=___ to process a set of volumes, multiple
	filters can be specified separated with commas.
EOH
	exit(0);
}

my $host = hostname;
logger("eql> running on host: $host");

my $group_name = $cfg{'default_eqlogic'};
if ( defined($opts{'g'}) )
{
	$group_name = $opts{'g'};
}

my $volume_name;
if ( defined($opts{'s'}) )
{
	$volume_name = $opts{'s'};
}

my $ddless_filter;
if ( defined($opts{'f'}) )
{
	$ddless_filter = $opts{'f'};
}

if ( !defined($volume_name) && !defined($ddless_filter) )
{
	logger_err("err> either specify [-f] or [-s volume]");
	exit 1;
}

luncopy_launcher($group_name, $volume_name, $ddless_filter);
exit 0;

#-------------------------------------------------------------------------------
# launch/parallelize luncopy
#-------------------------------------------------------------------------------
sub luncopy_launcher
{
	my ( $group_name, $volume_name, $ddless_filter ) = @_;

	#
	# we need to be able to process a single volume or multiples, the decision
	# of whether we are expanding a filter or processing a single volume is
	# based on the volume_name being defined.
	#
	if ( defined($volume_name) )
	{
		#
		# the filter is passed along to allow subsequent components to know
		# where to put the destination data
		#
		luncopy_volume($group_name, $volume_name, $ddless_filter);
	}
	else
	{
		#
		# connect with eql
		#
		my $eql;
		$eql = eql_connect($cfg{$group_name}{"ip"},
			$cfg{$group_name}{"user"},
			$cfg{$group_name}{"pass"});
		if ( !$eql)
		{
			logger("err> unable to connect to equallogic device");
			exit 1;
		}
		
		#
		# get volume meta data
		#
		$cfg{'verbose'} = 0;
		my $volumes = eql_volumes($eql);
		#eql_list_dump($volumes, "data> volumes", ("Name","Size"));
		$cfg{'verbose'} = 1;
		
		#
		# disconnect from eql
		#
		eql_disconnect($eql);
		
		logger("copy> filter for DDLESS=$ddless_filter");
		my @filters = split /,/,$ddless_filter;
		my @volumes;
		my %volume_filter;
		foreach my $volume_data (@$volumes)
		{
			my $volume_name = $$volume_data{'Name'};

			for my $filter (@filters)
			{
				if ( $$volume_data{'Description'} =~ /DDLESS=$filter/ )
				{
					logger("copy> adding volume $volume_name to this job with filter $filter");
					push @volumes, $volume_name;
					$volume_filter{$volume_name} = $filter;
				}
			}
		}
		
		#
		# sort them in order to distribute the work logically
		#
		@volumes = sort @volumes;

		#
		# use the parallelizer code, invoke myself via shell script to
		# forcably redirect stdout/err to a log file that is volume specific
		#
		my %args = ( 
			'group_name' => $group_name,
			'volume_filter' => \%volume_filter
		);
		parallelize($cfg{$group_name}{'concurrent_workers'}, \&luncopy_spawn, \%args, @volumes);
	}

	logger("copy> processing complete");
}

#-------------------------------------------------------------------------------
# parallalized luncopy
#-------------------------------------------------------------------------------
sub luncopy_spawn
{
	my($volume_name, %args) = @_;

	#
	# need to delay to not overrun the equallogic with too many concurrent
	# admin sessions (2 minutes)
	#	
	my $delay = rand(120);
	logger("copy> delaying worker by $delay seconds");
	sleep($delay);

	my $log_file = "$cfg{'log_directory'}/${group_name}_${volume_name}.log";
	my $group_name = $args{'group_name'};
	my $ddless_filter = $args{'volume_filter'}{$volume_name};
	
	my $cmd = "$0 -g $group_name -s $volume_name -f $ddless_filter> $log_file 2>&1";
	logger("copy> spawn $cmd");
	system($cmd);
	if ( $? != 0 )
	{
		logger_err("copy_err> cmd $cmd");
		logger_err("copy_err> child exit value ".($?>>8));
		return undef;
	}
	
	print "\n";
	print "------------------------------------------------------\n";
	print ">>> $log_file\n";
	print "------------------------------------------------------\n";
	print `cat $log_file | grep -i err`;
	print `cat $log_file | grep "source device size"`;
	print `cat $log_file | grep "changed segments"`;
	print `cat $log_file | grep "wrote"`;
	print `tail -1 $log_file`;
	print "\n";

	#
	# keep historical copies
	#
	my $hist_dir = "$cfg{'log_directory'}/${group_name}_${volume_name}.history";
	if ( ! -d $hist_dir)
	{
		mkdir($hist_dir);
	}
	my $stamp = zfs_zvol_timestamp();
	copy($log_file,"$hist_dir/${group_name}_${volume_name}_$stamp.log");
	
	return 1;
}

#-------------------------------------------------------------------------------
# luncopy (the core code!)
#-------------------------------------------------------------------------------
sub luncopy_volume
{
	my( $group_name, $volume_name, $ddless_filter ) = @_;

	logger("copy> source group: $group_name");
	logger("copy> source volume: $volume_name");
	logger("copy> ddless filter: $ddless_filter");

	if ( !defined($cfg{'eql_luncopy_map'}{$host} ) )
	{
		logger("err> unable to find host $host in cfg{'eql_luncopy_map'}");
		return undef;
	}

	my $expose_snapshots_to_ip = $cfg{'eql_luncopy_map'}{$host}{'expose_snapshots_to_ip'};
	logger("copy> snapshots exposed to: $expose_snapshots_to_ip");

	#
	# use either the default pool 
	#
	my $zfs_pool = $cfg{'eql_luncopy_map'}{$host}{'pool_default'};
	if ( defined($ddless_filter) )
	{
		$zfs_pool = $cfg{'eql_luncopy_map'}{$host}{'filter_pool_map'}{$ddless_filter};
	}
	my $zfs_vol = "${group_name}_${volume_name}_bk";
	my $zfs_opts = $cfg{'eql_luncopy_map'}{$host}{'pool_zfs_options'}{$zfs_pool};
	
	logger("copy> zpool: $zfs_pool");
	logger("copy> zvol: $zfs_vol");
	logger("copy> zfs_opts: $zfs_opts");
	
	#
	# connect with equallogic unit
	#
	my $eql = eql_connect($cfg{$group_name}{"ip"},
		$cfg{$group_name}{"user"},
		$cfg{$group_name}{"pass"});
	if ( !$eql)
	{
		logger("err> unable to connect to equallogic device");
		return undef;
	}
	
	#
	# create a snapshot on equallogic and generate the zvol timestamp
	#
	my $stamp = zfs_zvol_timestamp();
	my $snapshot_name = equallogic_smart_snapshot_create($eql, $group_name, $volume_name, $expose_snapshots_to_ip);
	if ( !defined($snapshot_name) )
	{
		logger("err> unable to create equallogic snapshot for $volume_name");
		return undef;
	}

	#
	# get snapshot iscsi target IQN
	#
	my $snapshot_data = eql_volume_select_snapshot_select_show($eql, $volume_name, $snapshot_name);
	my $source_target_iqn = $$snapshot_data{"iSCSI Name"};
	logger ("copy> source target IQN: $source_target_iqn");

	#
	# disconnect from equallogic, long running snapshots cause the communication
	# channel to close, therefore, we might as well close it now
	#
	eql_disconnect($eql);
	
	#
	# connect via SUN iscsi initiator to the equallogic unit
	#
	my %san_conn = (
		'san_ip' => $cfg{$group_name}{"ip"},
		'san_ip_port' => '3260',
		'san_target' => $source_target_iqn,
	);

	my $ccode = iscsiadm_connect(\%san_conn);
	if ( !defined($ccode) )
	{
		logger("err> unable to connect with equallogic via iscsi initiator");
		return undef;
	}
	
	#
	# the equallogic has one LUN per target - we do not need to enumerate
	# the LUNs then
	#
	my $device = $san_conn{'luns'}{'0'}{'device'};
	if ( !defined($device) )
	{
		logger("err> unable to find LUN 0 device name");
		return undef;
	}
	logger("copy> source device $device");

	my %device_info = sun_disk_info($device);
	logger("copy> device serial number ".$device_info{'inquiry-serial-no'});
	logger("copy> device size $device_info{'disk_size_bytes'} bytes ($device_info{'disk_size_gb'} GB)");

	#
	# provision zfs volume (please remember that the _bk suffix only applies
	# to the volume here not to the ddless signature file, since it is based on
	# reading the source disk data)
	#
	my $ccode = zfs_create($zfs_pool, $zfs_vol, 
		$device_info{'disk_size_bytes'}, $cfg{'ddless_blocksize_bytes'},
		$zfs_opts);
	if ( !defined($ccode) )
	{
		logger_err("zfs_err> access pool/volume/grow volume?");
		return undef;
	}

	#
	# check for volume growth, if so, recreate the checksum file using the
	# data from the current archive set (i.e. the set that was initially
	# of size x and just now via zfs_create got grown to size x+n.
	# This prevents ddless from re-creating a new ddless signature file when
	# the volume growths and hence it would otherwise re-write all of the data
	# over again - no - no - we just want the new data
	#
	# destination must not be the raw device, as that will cause write errors,
	# like perror: File too large, write failed.
	#
	my $checksum_file = "/$zfs_pool/ddless/${group_name}_${volume_name}.ddless";
	my $dest_dev = "/dev/zvol/dsk/$zfs_pool/$zfs_vol";
	if ( $ccode == 2 )
	{
		logger("copy> resize of target detected, recreating signature file");
		logger("copy> using target's device file...");
		my $cmd = "$cfg{'ddless'} -v -s $dest_dev -c $checksum_file -w 4";
		logger("copy> $cmd");
		system($cmd);
		if ( $? != 0 )
		{
			logger_err("ddless_err> cmd $cmd");
			logger_err("ddless_err> child exit value ".($?>>8));
			return undef;
		}	
	}
		
	#
	# execute ddless
	# The source_dev must change the slice 2 reference from the device
	# to partition 0. In SUN tradition slice 2 was an overlapping slice
	# that referred to the entire disk.
	#
	# we used to use the raw device for the destination but have found
	# as solaris release progressed that this can fail with write errors
	# on a new zvols - reverted back to dsk. Reading appears fine.
	#
	my $source_dev = $device;
	$source_dev =~ s/s2$/p0/;

	#
	# 2tb check (pre)
	#
	my $disk_bug = disk_read_2TB_no_part_bug_before_ddless($source_dev,
		$zfs_vol);
	if ( !defined($disk_bug) )
	{
		logger_err("err> device size problem encountered!!!");
		return undef;
	}

	my $cmd = "$cfg{'ddless'} -v -s $source_dev -t $dest_dev -c $checksum_file -w 4";
	logger("copy> $cmd");
	system($cmd);
	if ( $? != 0 )
	{
		#
		# flag this temporarily as an error - until this is more tested
		#
		logger_err("ddless_err> unexpected exit code, retrying $zfs_vol");
		
		logger("ddless> child exit value ".($?>>8));
		logger("ddless> retrying operation after 5 minutes...");
		sleep(5 * 60);

		logger("copy> $cmd");
		system($cmd);
		if ( $? != 0 )
		{
			logger_err("ddless_err> cmd $cmd");
			logger_err("ddless_err> child exit value ".($?>>8));
			logger_err("ddless_err> retry on $zfs_vol failure");
			return undef;
		}
		logger_err("ddless_err> retry on $zfs_vol successful");
	}

	#
	# 2tb check (post)
	#
	if ( $disk_bug == 2 )
	{
		disk_read_2TB_no_part_bug_after_ddless($dest_dev, $checksum_file, 
			$zfs_vol);
	}

	#
	# backup current ddless signature file (so that we can compare it to the
	# snapshot we took before ddless was run)
	#
	my $hist_dir = "/$zfs_pool/ddless/${group_name}_${volume_name}.history";
	if ( ! -d $hist_dir)
	{
		mkdir($hist_dir);
	}
	my $checksum_bk = "$hist_dir/${group_name}_${volume_name}_$stamp.ddless";
	copy($checksum_file, $checksum_bk);
	
	#
	# snapshot zfs volume
	#
	if ( !zfs_create_snapshot(undef, $zfs_pool, $zfs_vol, $stamp) )
	{
		logger_err("zfs_err> unable to establish snapshots");
		return undef;
	}
	
	#
	# disconnect iscsi initiator session from equallogic device
	#
	$ccode = iscsiadm_disconnect(\%san_conn);
	if ( !defined($ccode) )
	{
		logger_err("err> unable to disconnect from equallogic using iscsi initiator");
		return undef;
	}

	#
	# re-connect with equallogic unit
	#
	my $eql = eql_connect($cfg{$group_name}{"ip"},
		$cfg{$group_name}{"user"},
		$cfg{$group_name}{"pass"});
	if ( !$eql)
	{
		logger("err> unable to connect to equallogic device");
		return undef;
	}

	#
	# remove snapshot from equallogic
	#
	if ( !equallogic_smart_snapshot_remove($eql, 
		$group_name, $volume_name, $snapshot_name, $expose_snapshots_to_ip) )
	{
		logger("err> unable to remove equallogic snapshot $snapshot_name from $volume_name");
		return undef;
	}
}	


#-------------------------------------------------------------------------------
# equallogic snapshot create, re-use snapshot (in case of restarts)
#
# we keep a copy of the snapshot name on local disk, we need to see if we have
# stale snapshots. On return we provide the actual equallogic target's snapshot
# name
#-------------------------------------------------------------------------------
sub equallogic_smart_snapshot_create
{
	my ($eql, $group_name, $volume_name, $expose_snapshots_to_ip) = @_;
	
	my $snapshot_name;

	#
	# check if the volume has the ddless_snap disabled, i.e. use the second last
	# snapshot instead of creating your own snapshot at this point in time. The
	# second last snapshot is guaranteed to be closed by the host integration
	# toolkit in terms of the exchange databases.
	#
	my $ddless_snap = 1;
	my $volume_data = eql_volume_select_show($eql, $volume_name);
	if ( $$volume_data{'Description'} =~ /DDLESS_SNAP=0/ )
	{
		$ddless_snap = 0;
	}
	logger("info> ddless_snap=$ddless_snap"); 

	#
	# if we control the snapshot
	#
	if ( $ddless_snap )
	{
		#
		# open and read the snap state file for this volume
		#
		my $snap_state_file = "$cfg{'state_directory'}/$group_name-$volume_name.snapname";
		my $prev_snapshot_name;
		if ( open(FH, "<$snap_state_file") )
		{
			$prev_snapshot_name = <FH>;
			chomp($prev_snapshot_name);
			close(FH);
		}
		
		#
		# check if there is a possible previous valid snapshot
		#
		my $create_snap = 1;
		if ( defined($prev_snapshot_name) )
		{
			#
			# let's pretend we will use the previously defined snapshot...
			#
			logger("copy> found previously cached snapshot reference: $prev_snapshot_name");
			$snapshot_name = $prev_snapshot_name;
			
			#
			# get a list of snapshot names from the equallogic
			#
			my $snapshots = eql_volume_select_snapshot_show($eql, $volume_name);
			eql_list_dump($snapshots, "data> volume $volume_name snapshot", ("Name"));
			if ( eql_snapshot_lookup($snapshots, $snapshot_name) )
			{
				logger("info> found matching snapshot, using $snapshot_name");
				$create_snap = 0;
			}
			else
			{
				logger("info> looks like our snapshot name is stale, removing it, creating new snapshot");
				unlink $snap_state_file;
			}
		}

		#
		# create snapshot on equallogic
		#
		if ( $create_snap )
		{
			$snapshot_name = eql_volume_select_snapshot_create_now($eql, $volume_name);
			if ( !defined($snapshot_name) )
			{
				logger("err> unable to create eqlogic snapshot");
				return undef;
			}
		
			logger("info> created $snapshot_name, recording its name");
			if ( open(FH, ">$snap_state_file") )
			{
				print FH $snapshot_name;
				close(FH);
			}
		}
	}
	else
	{
		#
		# we have to use an existing snapshot that has been properly closed
		# by the equallogic intergration kit, do not choose the newest one, as
		# it could be still being closed, pick the one before that.
		#
		# NO - this caused issues with jobs starting late, like at 3:00am and
		# then we picked the one from 23:59 which was not properly closed.
		#
		# So, at this time, we expect at least 4 snapshots and we pick the
		# third one from the start (the array is zero based!)
		#
		logger("info> using existing snapshots on equallogic...");
		my $snapshots = eql_volume_select_snapshot_show($eql, $volume_name);

		my $snap_count = scalar(@{$snapshots});
		my $snapshot_connections = 0;
		my $snapshot_status;
		if ( $snap_count >= 4 )
		{
			#my $snapshot = @{$snapshots}[$snap_count-2];
			my $snapshot = @{$snapshots}[3];
			$snapshot_name = $$snapshot{'Name'};
			$snapshot_connections = $$snapshot{'Connections'};
			$snapshot_status = $$snapshot{'Status'};
		}
		foreach my $snapshot (@{$snapshots})
		{
			my $using = ($$snapshot{'Name'} eq $snapshot_name ? "using it":"");
			logger("info> snapshot: $$snapshot{'Name'} $using");
		}
		logger("info> chosen snapshot: $snapshot_name");
		if ( !defined($snapshot_name) )
		{
			logger("err> unable to find eqlogic snapshot");
			return undef;
		}

		logger("info> connections: $snapshot_connections, status: $snapshot_status");
		if ( $snapshot_connections != 0 )
		{
			logger("err> the snapshot $snapshot_name has $snapshot_connections connections open, refusing it");
			return undef;
		}
		if ( $snapshot_status ne 'online' )
		{
			logger("err> the snapshot $snapshot_name is expected to be online and it is: $snapshot_status");
			return undef;
		}
	}

	#
	# modify snapshot access control, such that luncopy can see it
	#
	my $acls = eql_volume_select_show_access($eql, $volume_name);
	eql_list_dump($acls, "data> volume $volume_name acl",("ID","Ipaddress"));
	if ( eql_volume_access_lookup($acls, "Ipaddress", $expose_snapshots_to_ip) )
	{
		logger("info> found matching acl, keeping it");
	}
	else
	{
		logger("info> creating acl");
		my $ccode = eql_volume_select_access_create($eql,$volume_name,
			"ipaddress $expose_snapshots_to_ip apply-to snapshot");
		if ( !defined($ccode) )
		{
			logger("err> unable to create eqlogic acl");
			return undef;
		}
	}

	#
	# if we control the snapshot
	#
	if ( $ddless_snap )
	{
		#
		# make snapshot online
		#
		my $ccode = eql_volume_select_snapshot_select_operation(
			$eql,$volume_name,$snapshot_name,"online");
		if ( !defined($ccode) )
		{
			logger("err> unable to bring snapshot online");
			return undef;
		}
	}

	return $snapshot_name;
}

#-------------------------------------------------------------------------------
# remove snapshot from equallogic device
#-------------------------------------------------------------------------------
sub equallogic_smart_snapshot_remove
{
	my ($eql, $group_name, $volume_name, $snapshot_name, $expose_snapshots_to_ip) = @_;
	
	#
	# check if the volume has the ddless_snap disabled, i.e. use the second last
	# snapshot instead of creating your own snapshot at this point in time. The
	# second last snapshot is guaranteed to be closed by the host integration
	# toolkit in terms of the exchange databases.
	#
	my $ddless_snap = 1;
	my $volume_data = eql_volume_select_show($eql, $volume_name);
	if ( $$volume_data{'Description'} =~ /DDLESS_SNAP=0/ )
	{
		$ddless_snap = 0;
	}
	logger("info> ddless_snap=$ddless_snap"); 

	#
	# if we control the snapshot
	#
	if ( $ddless_snap )
	{
		#
		# make snapshot offline
		#
		my $ccode = eql_volume_select_snapshot_select_operation(
			$eql,$volume_name,$snapshot_name,"offline");
		if ( !defined($ccode) )
		{
			logger("err> unable to bring snapshot offline");
			return undef;
		}
	}
	
	#
	# remove acl
	#
	my $acls = eql_volume_select_show_access($eql, $volume_name);
	eql_list_dump($acls, "data> volume $volume_name acl",("ID","Ipaddress"));
	my $acl = eql_volume_access_lookup($acls, "Ipaddress", $expose_snapshots_to_ip);
	if ( defined($acl) )
	{
		logger("copy> found matching acl ".$$acl{'ID'}.", removing it");
		my $ccode = eql_volume_select_access_delete($eql,$volume_name,$$acl{'ID'});
		if ( !defined($ccode) )
		{
			logger("err> unable to delete eqlogic acl");
			return undef;
		}
	}
	
	#
	# if we control the snapshot
	#
	if ( $ddless_snap )
	{
		#
		# remove snapshot
		#
		my $ccode = eql_volume_select_snapshot_delete($eql,$volume_name, $snapshot_name);
		if ( !defined($ccode) )
		{
			logger("err> unable to remove snapshot");
			return undef;
		}
	}
	
	return 1;
}


