#!/usr/bin/perl

# Equallogic Restore from ZFS, by Steffen Plotner, 2009

use strict;
use Data::Dumper;
use Getopt::Std;
use File::Copy;
use Sys::Hostname;
use POSIX qw(ceil floor);

use FindBin;
use lib "$FindBin::Bin/lib";
use lib "$FindBin::Bin/etc";

# vendor provided
use EqlScript;

# amherst provided
use common;
use config;
use eqlutil;
use SunIscsiInitiator;
use SunDiskInfo;
use Sun2Tb;

$cfg{'verbose'} = 1;

#
# handle command line options
#
my %opts;
getopts('hg:s:t:',\%opts);
if ( defined($opts{'h'}) )
{
	print <<EOH;
$0 [-h] -s zvol [-g eqlogicname] -t volume

-h	help

-s	source zfs volume (zpool/zvol or zpool/zvol\@snapshot)
-g	equallogic name (must be defined in eql.conf, default $cfg{'default_eqlogic'})
-t	target equallogic volume (must NOT exist)
EOH
	exit(0);
}

my $host = hostname;
logger("rest> running on host: $host");

my $group_name = $cfg{'default_eqlogic'};
if ( defined($opts{'g'}) )
{
	$group_name = $opts{'g'};
}

my $source_volume;
if ( defined($opts{'s'}) )
{
	$source_volume = $opts{'s'};
}

my $target_volume;
if ( defined($opts{'t'}) )
{
	$target_volume = $opts{'t'};
}

my $checksum_file = '/dev/null';

if ( !defined($source_volume) || !defined($target_volume) )
{
	logger_err("err> requires -s and -t");
	exit(1);
}

eql_restore($source_volume, $target_volume, $group_name, $checksum_file);
exit(0);

#-------------------------------------------------------------------------------
# equallogic restore code
#-------------------------------------------------------------------------------
sub eql_restore
{
	my ($source_volume, $target_volume, $group_name, $checksum_file) = @_;
		
	logger("rest> source volume: $source_volume");
	logger("rest> target volume: $target_volume");
	logger("rest> group name: $group_name");
	logger("rest> checksum file: $checksum_file");

	#
	# determine the source volume's device size
	#
	my $source_dev = "/dev/zvol/dsk/$source_volume";
	logger("rest> source device: $source_dev");
	if ( ! -e "$source_dev" )
	{
		logger("err> unable to find source_dev: $source_dev");
		exit(1);
	}

	#
	# the equallogic requires the input of the size in units of MB
	# use the ceiling function to round up the disk size (besides that
	# we need to add 1GB to make sure the equallogic's size will
	# work with ddless (ddless reqires the target to be >= than source)
	#
	my $source_size_bytes = sun_disk_size($source_dev);
	logger("rest> source size: $source_size_bytes bytes");
	my $source_size_mb = ceil($source_size_bytes / 1024 / 1024);
	logger("rest> source size: $source_size_mb MB");
	my $target_proposed_size_mb = $source_size_mb + 1024;
	logger("rest> target proposed size: $target_proposed_size_mb MB (safe)");

	# source (zfs)
	# 12897484800

	# target (eql)
	# 12880788480	12300 MB
	# 12889013760	12301 MB
	# 12889013760	12310 MB

	#
	# connect with equallogic unit
	#
	my $eql = eql_connect($cfg{$group_name}{"ip"},
		$cfg{$group_name}{"user"},
		$cfg{$group_name}{"pass"});
	if ( !$eql)
	{
		logger("err> unable to connect to equallogic device");
		return undef;
	}

	#
	# attempt to create volume (we will not overwrite an existing one)
	#
	# if you think you really want that because you are using the checksum
	# file and you are doing a restore this way, then comment out this error
	# condition - you have been warned!
	#
	if ( !eql_volume_create($eql, $target_volume, $target_proposed_size_mb))
	{
		logger("err> unable to create volume");
		return undef;
	}

	#
	# ok, this should probably be called expose_device_to_ip - however
	#
	my $expose_snapshots_to_ip = $cfg{'eql_luncopy_map'}{$host}{'expose_snapshots_to_ip'};
	logger("rest> snapshots exposed to: $expose_snapshots_to_ip");

	#
	# modify access control, such that we can see the target device
	#
	my $acls = eql_volume_select_show_access($eql, $target_volume);
	eql_list_dump($acls, "data> volume $target_volume acl",("ID","Ipaddress"));
	if ( eql_volume_access_lookup($acls, "Ipaddress", $expose_snapshots_to_ip) )
	{
		logger("info> found matching acl, keeping it");
	}
	else
	{
		logger("info> creating acl");
		my $ccode = eql_volume_select_access_create($eql,$target_volume,
			"ipaddress $expose_snapshots_to_ip apply-to volume");
		if ( !defined($ccode) )
		{
			logger("err> unable to create eqlogic acl");
			return undef;
		}
	}
	
	#
	# get iscsi target IQN from volume
	#
	my $target_volume_data = eql_volume_select_show($eql, $target_volume);
	my $target_iqn = $$target_volume_data{"iSCSI Name"};
	logger ("rest> target IQN: $target_iqn");

	#
	# disconnect from equallogic, long running snapshots cause the communication
	# channel to close, therefore, we might as well close it now
	#
	eql_disconnect($eql);


	#
	# connect via SUN iscsi initiator to the equallogic unit
	#
	my %san_conn = (
		'san_ip' => $cfg{$group_name}{"ip"},
		'san_ip_port' => '3260',
		'san_target' => $target_iqn,
	);

	my $ccode = iscsiadm_connect(\%san_conn);
	if ( !defined($ccode) )
	{
		logger("err> unable to connect with equallogic via iscsi initiator");
		return undef;
	}
	
	#
	# the equallogic has one LUN per target - we do not need to enumerate
	# the LUNs then
	#
	my $target_dev = $san_conn{'luns'}{'0'}{'device'};
	if ( !defined($target_dev) )
	{
		logger("err> unable to find LUN 0 device name");
		return undef;
	}
	logger("rest> target device $target_dev");

	#
	# The target_dev must change the slice 2 reference from the device
	# to partition 0. In SUN tradition slice 2 was an overlapping slice
	# that referred to the entire disk.
	#
	$target_dev =~ s/s2$/p0/;
	logger("rest> target device $target_dev");

	my %device_info = sun_disk_info($target_dev);
	logger("rest> target device serial number ".$device_info{'inquiry-serial-no'});
	logger("rest> target device size $device_info{'disk_size_bytes'} bytes ($device_info{'disk_size_gb'} GB)");

	#
	# 2tb check (pre, post step not needed during restore)
	#
	my $disk_bug = disk_read_2TB_no_part_bug_before_ddless($target_dev);
	if ( !defined($disk_bug) )
	{
		logger_err("err> device size problem encountered!!!");
		return undef;
	}

	#
	# restore data using ddless
	#
	my $cmd = "$cfg{'ddless'} -v -s $source_dev -t $target_dev -c $checksum_file -w 4";
	logger("rest> $cmd");
	system($cmd);
	if ( $? != 0 )
	{
		logger_err("ddless_err> cmd $cmd");
		logger_err("ddless_err> child exit value ".($?>>8));
		return undef;
	}	

	
	#
	# disconnect iscsi initiator session from equallogic device
	#
	$ccode = iscsiadm_disconnect(\%san_conn);
	if ( !defined($ccode) )
	{
		logger("err> unable to disconnect from equallogic using iscsi initiator");
		return undef;
	}

	#
	# re-connect with equallogic unit
	#
	my $eql = eql_connect($cfg{$group_name}{"ip"},
		$cfg{$group_name}{"user"},
		$cfg{$group_name}{"pass"});
	if ( !$eql)
	{
		logger("err> unable to connect to equallogic device");
		return undef;
	}

	#
	# remove access control entry created earlier
	#
	my $acls = eql_volume_select_show_access($eql, $target_volume);
	eql_list_dump($acls, "data> volume $target_volume acl",("ID","Ipaddress"));
	my $acl = eql_volume_access_lookup($acls, "Ipaddress", $expose_snapshots_to_ip);
	if ( defined($acl) )
	{
		logger("rest> found matching acl ".$$acl{'ID'}.", removing it");
		my $ccode = eql_volume_select_access_delete($eql,$target_volume,$$acl{'ID'});
		if ( !defined($ccode) )
		{
			logger("err> unable to delete eqlogic acl");
			return undef;
		}
	}
	
	logger("rest> processing complete");
	return 1;
}
