#!/usr/bin/perl

# IET luncopy to ZFS, by Steffen Plotner, 2009, 2010

use strict;
use Data::Dumper;
use Getopt::Std;
use File::Copy;
use Sys::Hostname;
use Digest::MD5;

use FindBin;
use lib "$FindBin::Bin/lib";
use lib "$FindBin::Bin/etc";

# amherst provided
use common;
use config;
use ietutil;
use SunIscsiInitiator;
use SunDiskInfo;
use SunZfs;
use Sun2Tb;

$cfg{'verbose'} = 1;

my $max_ddless_check = 4;

#
# handle command line options
#
my %opts;
getopts('hais:l',\%opts);
if ( defined($opts{'h'}) )
{
	print <<EOH;
$0 [-h] -a [-i] | -s vg/lv [-i] | -l

-h	help

-a	process all volumes
-s	process single vg/lv

-i	integrity check of last vg/lv luncopy operation

-l	show a listing of candidate volumes
EOH
	exit(0);
}

my $host = hostname;
logger("iet> running on host: $host");

my $integrity_check = 0;
if ( defined($opts{'i'}) )
{
	$integrity_check = 1;
}

#
# list volumes
#
if ( defined($opts{'l'}) )
{
	$cfg{'verbose'} = 0;
	my %volumes = iet_volumes();
	#print Dumper \%volumes;
	foreach my $volume (sort keys %volumes)
	{
		print "$volume\n";
	}
}
#
# process a single volume
#
elsif ( defined($opts{'s'}) )
{
	my $backup_vol = $opts{'s'};
	my %volumes = iet_volumes();
	#print Dumper \%volumes;
	my $found = 0;
	foreach my $volume (sort keys %volumes)
	{
		if ( $volume eq $backup_vol )
		{
			resource_luncopy_volume($backup_vol, $volumes{$backup_vol});
			$found = 1;
		}
	}
	if ( !$found )
	{
		print "err> volume $backup_vol not found, use -l to get a listing\n";
		exit(1);
	}
}
#
# process all volumes
#
elsif ( defined($opts{'a'}) )
{
	$cfg{'verbose'} = 0;
	my %volumes = iet_volumes();
	$cfg{'verbose'} = 1;
	#print Dumper \%volumes;
	my @volumes_selection = work_luns(sort keys %volumes);
	foreach my $volume (@volumes_selection)
	{
		#if ( $volume eq "vg_satabeast8_vol0/lv_grey1_in" )
		#{
			#logger("WORK ON: $volume");
			luncopy_spawn($volume, $volumes{$volume});
		#}
	}
	logger("copy> processing complete");
}

exit 0;

#-------------------------------------------------------------------------------
# determine a set of luns to work on, either ALL, or only a limited set of luns
# that we would use during integrity check
#-------------------------------------------------------------------------------
sub work_luns
{
	my (@luns) = @_;
	
	#
	# if no integrity, then return all volumes
	#
	if ( $integrity_check == 0 )
	{
		return @luns;
	}
	
	my $state_file = "$cfg{'state_directory'}/ietcheck.dat";

	#
	# state: first time, create a new file
	#
	if ( ! -e "$state_file" )
	{
		logger("copy> no existing state file, creating new one");
		if ( !open(FH,">$state_file") )
		{
			logger("copy_err> unable to create file $state_file: $?");
			return undef;
		}
		close(FH);
	}

	#
	# read existing state
	#
	my @state_luns;
	if ( !open(FH, "<$state_file") )
	{	
		logger("copy_err> unable to open file $state_file: $?");
		return undef;
	}
	while(<FH>)
	{
		my $line = $_;
		chomp($line);
		if ( $line ne "" )
		{
			push @state_luns, $line;
		}
	}
	close(FH);

	#
	# how many zvols do we have?
	#
	my $lun_count = scalar(@state_luns);
	logger("copy> read $lun_count volumes");

	#
	# if there are no vols, then populate the state using all vols from system
	#
	if ( $lun_count == 0 )
	{
		@state_luns = @luns;
	}
	
	#
	# prepare n luns for work
	#
	my @work_luns;
	my $did_work = 0;
	my $max_work = $max_ddless_check;
	while ( $did_work < $max_work )
	{
		my $lun = pop @state_luns;
		if ( defined($lun) )
		{
			push @work_luns, $lun;
		}
		
		$did_work++;
	}

	#
	# write out the state
	#
	if ( !open(FH, ">$state_file") )
	{
		logger("copy_err> unable to open file $state_file: $?");
		return undef;
	}
	foreach my $lun (@state_luns)
	{
		print FH "$lun\n";
	}
	close(FH);

	#
	# debug the csc1_gfs case (this breaks on zfs-archive2 of course)
	#
	#if ( $host eq 'zfs-archive1' )
	#{
	#	push @work_luns, "vg_satabeast8_vol0/lv_csc1_gfs_vol00";
	#}
	
	return @work_luns;
}

#-------------------------------------------------------------------------------
# luncopy spawn (track log files)
#-------------------------------------------------------------------------------
sub luncopy_spawn
{
	my ($vg_lv, $volume) = @_;

	my $iet_host = $$volume{'iet_host'};
	my $iet_lv = $$volume{'iet_lv'};

	my $zfs_prefix = $cfg{'iet_cluster'}{$iet_host}{'zfs_prefix'};

	my $log_file = "$cfg{'log_directory'}/${zfs_prefix}_${iet_lv}.log";
	
	my $integrity_args;
	if ( $integrity_check )
	{
		$integrity_args = "-i";
	}
	
	my $cmd = "$0 -s $vg_lv $integrity_args > $log_file 2>&1";
	logger("copy> spawn $cmd");
	system($cmd);
	if ( $? != 0 )
	{
		logger_err("copy_err> cmd $cmd");
		logger_err("copy_err> child exit value ".($?>>8));
		return undef;
	}
	
	print "\n";
	print "------------------------------------------------------\n";
	print ">>> $log_file\n";
	print "------------------------------------------------------\n";
	print `cat $log_file | grep -i err`;
	print `cat $log_file | grep "ddmap_state"`;
	print `cat $log_file | grep "source device size"`;
	print `cat $log_file | grep "changed segments"`;
	print `cat $log_file | grep "wrote"`;
	print `tail -1 $log_file`;
	print "\n";

	#
	# keep historical copies
	#
	my $hist_dir = "$cfg{'log_directory'}/${zfs_prefix}_${iet_lv}.history";
	if ( ! -d $hist_dir)
	{
		mkdir($hist_dir);
	}
	my $stamp = zfs_zvol_timestamp();
	copy($log_file,"$hist_dir/${zfs_prefix}_${iet_lv}_$stamp.log");
	
	return 1;
}

#-------------------------------------------------------------------------------
# iet luncopy (with resource lock and timeout, we cannot allow
# job overruns in case we have all LUNs set to DDMAP_DIRTY, for example)
#-------------------------------------------------------------------------------
sub resource_luncopy_volume
{
	my ($vg_lv, $volume) = @_;

	my $iet_lv = $$volume{'iet_lv'};
	my $iet_vg = $$volume{'iet_vg'};

	#
	# lock (with timeout)
	#
	my $resource = $cfg{'state_directory'}."/iet_target_${iet_vg}_${iet_lv}";
	$SIG{ALRM} = sub { die "timeout" };
	my $lock;
	eval
	{
		alarm(30);
		$lock = resource_lock($resource);
		alarm(0);
	};
	if ($@)
	{
		#
		# if we don't acquire the lock in 30 seconds, alarm will die
		# with a timeout result, which we catch here
		#
		if ( $@ =~ /timeout/ )
		{
			logger_err("copy_err> unable to obtain resource lock in a timely fashion");
			exit(1);
		}
		alarm(0);
	}
	if ( !defined($lock) )
	{
		logger_err("copy_err> unable to obtain resource lock $resource");
		return undef;
	}

	#
	# do work
	#
	luncopy_volume($vg_lv, $volume);
	
	#
	# unlock
	#
	resource_unlock($lock, $resource);
}

#-------------------------------------------------------------------------------
# iet luncopy
#-------------------------------------------------------------------------------
sub luncopy_volume
{
	my ($vg_lv, $volume) = @_;
	
	my $iet_host = $$volume{'iet_host'};
	my $iet_lv = $$volume{'iet_lv'};
	my $iet_vg = $$volume{'iet_vg'};
	my $zpool = $$volume{'dest_pool'};
	
	logger("copy> source host: $iet_host vg: $iet_vg lv: $iet_lv dest_zpool: $zpool");
	
	#
	# iet: note that we do NOT include the volume group in the name, which means
	# you cannot and should not ever have the same logical volume on the remote
	# side on 2 different volume groups, however, here it allows us to more
	# easily see the logical volumes like: iscsi4_lv_blackboard_bk.
	#
	my $zfs_prefix = $cfg{'iet_cluster'}{$iet_host}{'zfs_prefix'};
	my $zfs_pool = $zpool;
	my $zfs_vol = "${zfs_prefix}_${iet_lv}_bk";
	my $zfs_opts = $cfg{'iet_luncopy_map'}{$host}{"zfs_options"};
	
	logger("copy> zpool: $zfs_pool");
	logger("copy> zvol: $zfs_vol");
	logger("copy> zfs_opts: $zfs_opts");

	my $initiator_name = iscsiadm_initiator_name();
	logger("copy> initiator name: $initiator_name");
	
	#
	# we expect the code in this section to assign the meta
	# data about this snapshot, such as the target name
	# the ddmap state of the snapshot and the ddmap filename
	#
	my $stamp = zfs_zvol_timestamp();
	my %snap_meta = iet_lv_snap($iet_host, $vg_lv, $initiator_name, $integrity_check);
	if ( !defined(%snap_meta) )
	{
		logger_err("err> no meta data received");
		return undef;
	}

	#
	# at this point the snap_meta has been filled out, we
	# can verify the state of the ddmap and proceed with the copy
	#
	my $state_file = "/$zfs_pool/ddless/${zfs_prefix}_${iet_lv}.state";
	my $ddmap_file = "/$zfs_pool/ddless/${zfs_prefix}_${iet_lv}.ddmap";
	unlink($state_file);
	unlink($ddmap_file);
	
	if ( ! open(FH, ">$state_file") )
	{
		logger_err("err> unable to open $state_file: $!");
		return undef;
	}
	print FH "$snap_meta{'ddmap_state'}\n";
	close FH;
	
	#
	# decide if we have a clean or dirty state
	#
	my $ddless_ddmap_arg="";
	if ( $snap_meta{'ddmap_state'} eq 'DDMAP_CLEAN' )
	{
		logger("copy> found ddmap state CLEAN, using ddmap: $ddmap_file");
		$ddless_ddmap_arg = "-m $ddmap_file";

		#
		# fetch remote ddmap
		#
		if ( !defined(iet_ddmap_fetch($iet_host, $snap_meta{'ddmap_file'}, $ddmap_file)) )
		{
			logger_err("err> ddmap remote copy failed");
			return undef;
		}
	}
	else
	{
		#
		# this is not really an error, however, we would like to know
		# when this happens!
		#
		logger("copy_err> $iet_host: $iet_vg/$iet_lv: found ddmap state: $snap_meta{'ddmap_state'}, not using ddmap");
	}

	#
	# connect via SUN iscsi initiator to the iet unit
	#
	my %san_conn = (
		'san_ip' => $cfg{'iet_cluster'}{$iet_host}{"discovery_ip"},
		'san_ip_port' => '3260',
		'san_target' => $snap_meta{"target"},
	);

	my $ccode = iscsiadm_connect(\%san_conn);
	if ( !defined($ccode) )
	{
		logger("err> unable to connect with iet via iscsi initiator");
		return undef;
	}

	#
	# iet has support for multiple LUNs per target, however this job
	# only uses one
	#
	my $device = $san_conn{'luns'}{'0'}{'device'};
	if ( !defined($device) )
	{
		logger("err> unable to find LUN 0 device name");
		return undef;
	}

	#
	# with iet, the source device appears to just end with a the slice
	# delimiter and no slice number (strange). The problem comes from:
	#
	# scsiadm list target -S
	#
	# OS Device Name: /dev/rdsk/c0t494554000000000044534B2D313030353038323230373431d0s
	#
	logger("copy> source device $device");
	$device =~ s/s$/s2/;
	logger("copy> source device $device (named adjusted)");

	my %device_info = sun_disk_info($device);
	logger("copy> device serial number ".$device_info{'inquiry-serial-no'});
	logger("copy> device size $device_info{'disk_size_bytes'} bytes ($device_info{'disk_size_gb'} GB)");

	#
	# if the zfs destination volume does not exist at this point, then
	# clear the ddmap setting such that we are not going to use it - i.e. we
	# have to read the entire source disk
	#
	my $cmd = "zfs list -H -o name $zfs_pool/$zfs_vol";
	logger("copy> $cmd");
	system($cmd);
	if ( $? != 0 )
	{
		logger("copy> ignoring ddmap since the disk has been newly created");
		$ddless_ddmap_arg = "";
	}
	
	#
	# provision zfs volume (please remember that the _bk suffix only applies
	# to the volume here not to the ddless signature file, since it is based on
	# reading the source disk data)
	#
	my $ccode = zfs_create($zfs_pool, $zfs_vol, 
		$device_info{'disk_size_bytes'}, $cfg{'ddless_blocksize_bytes'},
		$zfs_opts);
	if ( !defined($ccode) )
	{
		logger_err("zfs_err> access pool/volume/grow volume?");
		return undef;
	}

	#
	# check for volume growth, if so, recreate the checksum file using the
	# data from the current archive set (i.e. the set that was initially
	# of size x and just now via zfs_create got grown to size x+n.
	# This prevents ddless from re-creating a new ddless signature file when
	# the volume growths and hence it would otherwise re-write all of the data
	# over again - no - no - we just want the new data
	#
	# destination must not be the raw device, as that will cause write errors,
	# like perror: File too large, write failed.
	#
	my $checksum_file = "/$zfs_pool/ddless/${zfs_prefix}_${iet_lv}.ddless";
	my $dest_dev = "/dev/zvol/dsk/$zfs_pool/$zfs_vol";
	if ( $ccode == 2 )
	{
		logger("copy> resize of target detected, recreating signature file");
		logger("copy> using target's device file...");
		my $cmd = "$cfg{'ddless'} -v -s $dest_dev -c $checksum_file -w 4";
		logger("copy> $cmd");
		system($cmd);
		if ( $? != 0 )
		{
			logger_err("ddless_err> cmd $cmd");
			logger_err("ddless_err> child exit value ".($?>>8));
			return undef;
		}
		
		#
		# the ddmap cannot be used if the disk was resized. The added disk
		# region is properly reflected in terms of actually written blocks,
		# but that region COULD have other random data that we would not know
		# about - we want the exact data - so ignore the ddmap on disk resize
		#
		logger("copy> ignoring ddmap since the disk has been resized");
		$ddless_ddmap_arg = "";
	}

	#
	# execute ddless (uses ddmap if the map is clean, otherwise not)
	#
	# The source_dev must change the slice 2 reference from the device
	# to partition 0. In SUN tradition slice 2 was an overlapping slice
	# that referred to the entire disk.
	#
	# we used to use the raw device for the destination but have found
	# as solaris release progressed that this can fail with write errors
	# on a new zvols - reverted back to dsk. Reading appears fine.
	#
	my $source_dev = $device;
	$source_dev =~ s/s2$/p0/;

	#
	# 2tb check (pre)
	#
	my $disk_bug = disk_read_2TB_no_part_bug_before_ddless($source_dev,
		$zfs_vol);
	if ( !defined($disk_bug) )
	{
		logger_err("err> device size problem encountered!!!");
		return undef;
	}

	#
	# ddless execute
	#
	my $cmd = "$cfg{'ddless'} -v -s $source_dev $ddless_ddmap_arg -t $dest_dev ".
		"-c $checksum_file -w 4";
	logger("copy> $cmd");
	system($cmd);
	if ( $? != 0 )
	{
		#
		# flag this temporarily as an error - until this is more tested
		#
		logger_err("ddless_err> unexpected exit code, retrying $zfs_vol");
		
		logger("ddless> child exit value ".($?>>8));
		logger("ddless> retrying operation after 5 minutes...");
		sleep(5 * 60);

		logger("copy> $cmd");
		system($cmd);
		if ( $? != 0 )
		{
			logger_err("ddless_err> cmd $cmd");
			logger_err("ddless_err> child exit value ".($?>>8));
			logger_err("ddless_err> retry on $zfs_vol failure");
			return undef;
		}
		logger_err("ddless_err> retry on $zfs_vol successful");
	}

	#
	# 2tb check (post)
	#
	if ( $disk_bug == 2 )
	{
		disk_read_2TB_no_part_bug_after_ddless($dest_dev, $checksum_file, 
			$zfs_vol);
	}

	#
	# if we are doing an integrity check, we now need to use the 
	# snap_meta{'ddless_md5'} md5 string and compare that to the md5 string
	# of the current $checksum_file
	#
	if ( $integrity_check )
	{
		logger("integrity> remote ddless_md5: $snap_meta{'ddless_md5'}");
		
		if ( !open(FH, "<$checksum_file") )
		{
			logger_err("copy_err> unable to open: $checksum_file");
			return undef;
		}
		binmode(FH);
		my $checksum_file_md5 = Digest::MD5->new->addfile(*FH)->hexdigest();
		close(FH);
		logger("integrity> local  ddless_md5: $checksum_file_md5");
		
		my $checksum_status;
		if ( $snap_meta{'ddless_md5'} eq $checksum_file_md5 )
		{
			$checksum_status = "checksum ok";
			logger("integrity> $iet_host: $iet_vg/$iet_lv: $checksum_status");
		}
		else
		{
			$checksum_status = "checksum mismatch";
			logger_err("integrity_err> $iet_host: $iet_vg/$iet_lv: $checksum_status");
			
			#
			# flag remote volume as dirty
			#
			if ( !defined(iet_lv_dirty($iet_host, $vg_lv)) )
			{
				logger_err("err> unable to flag volume as dirty");
				return undef;
			}
		}
		
		#
		# record the integrity checksum result
		#
		my $ietlog_file = "$cfg{'log_directory'}/iet_checksum.log";
		my $stamp = date_time_string();
		open(FH, ">>$ietlog_file");
		print FH "$stamp - $iet_host: $iet_vg/$iet_lv: $checksum_status\n";
		close(FH);
		
	}

	#
	# backup current ddless signature file (so that we can compare it to the
	# snapshot we took before ddless was run)
	#
	my $hist_dir = "/$zfs_pool/ddless/${zfs_prefix}_${iet_lv}.history";
	if ( ! -d $hist_dir)
	{
		mkdir($hist_dir);
	}
	my $checksum_bk = "$hist_dir/${zfs_prefix}_${iet_lv}_$stamp.ddless";
	copy($checksum_file, $checksum_bk);
	my $ddmap_bk = "$hist_dir/${zfs_prefix}_${iet_lv}_$stamp.ddmap";
	copy($ddmap_file, $ddmap_bk);
	my $state_bk = "$hist_dir/${zfs_prefix}_${iet_lv}_$stamp.state";
	copy($state_file, $state_bk);

	#
	# snapshot zfs volume
	#
	if ( !zfs_create_snapshot(undef, $zfs_pool, $zfs_vol, $stamp) )
	{
		logger_err("zfs_err> unable to establish snapshots");
		return undef;
	}

	#
	# disconnect iscsi initiator session from iet device
	#
	$ccode = iscsiadm_disconnect(\%san_conn);
	if ( !defined($ccode) )
	{
		logger_err("err> unable to disconnect from iet using iscsi initiator");
		return undef;
	}
	
	#
	# we MUST re-evaluate the iet_host because the snapshot target
	# might have moved while we were busy copying data
	#
	my %volumes = iet_volumes();
	if ( $iet_host ne $volumes{$vg_lv}{'iet_host'} )
	{
		logger("copy> iet_host has changed from $iet_host to $volumes{$vg_lv}{'iet_host'}");
		$iet_host = $volumes{$vg_lv}{'iet_host'};
	}

	#
	# remove iet snapshot and target
	#
	if ( !defined(iet_lv_snapremove($iet_host, $vg_lv, $initiator_name)) )
	{
		logger_err("err> unable to remove snapshot");
		return undef;
	}
}



