#!/usr/bin/perl

# iSCSI Enterprise Target restore from ZFS, by Steffen Plotner, 2011

use strict;
use Data::Dumper;
use Getopt::Std;
use File::Copy;
use Sys::Hostname;
use POSIX qw(ceil floor);

use FindBin;
use lib "$FindBin::Bin/lib";
use lib "$FindBin::Bin/etc";

# amherst provided
use common;
use config;
use ietutil;
use SunIscsiInitiator;
use SunDiskInfo;
use Sun2Tb;

$cfg{'verbose'} = 1;

#
# handle command line options
#
my %opts;
getopts('hg:s:t:',\%opts);
if ( defined($opts{'h'}) )
{
	print <<EOH;
$0 [-h] -s zvol -g volumegroup -t volume

-h	help

-s	source zfs volume (zpool/zvol or zpool/zvol\@snapshot)
-g	volume group (the host is determined by querying members of 'iet_cluster'
-t	target iet volume (must NOT exist)
EOH
	exit(0);
}

my $host = hostname;
logger("rest> running on host: $host");

my $source_volume;
if ( defined($opts{'s'}) )
{
	$source_volume = $opts{'s'};
}

my $target_vg;
if ( defined($opts{'g'}) )
{
	$target_vg = $opts{'g'};
}

my $target_volume;
if ( defined($opts{'t'}) )
{
	$target_volume = $opts{'t'};
}

my $checksum_file = '/dev/null';

if ( !defined($source_volume) || !defined($target_vg) || !defined($target_volume) )
{
	logger_err("err> requires -s, -g and -t");
	exit(1);
}

iet_restore($source_volume, $target_vg, $target_volume, $checksum_file);
exit(0);

#-------------------------------------------------------------------------------
# iet restore code
#-------------------------------------------------------------------------------
sub iet_restore
{
	my ($source_volume, $target_vg, $target_volume, $checksum_file) = @_;
		
	logger("rest> source volume: $source_volume");
	logger("rest> target vg: $target_vg");
	logger("rest> target volume: $target_volume");
	logger("rest> checksum file: $checksum_file");

	#
	# enumerate iet cluster hosts and search for the target vg
	#
	my $iet_host;
	foreach my $iet_host_search( keys %{$cfg{'iet_cluster'}} )
	{
		logger("rest> searching for target vg: $target_vg on host: $iet_host");
		
		# have the host implement vg check and return word "ok" if it is found
		# exit 1 is for fatal errors only
		
		if ( iet_vgcheck($iet_host_search, $target_vg) )
		{
			$iet_host = $iet_host_search;
			last;
		}
	}
	
	if ( ! defined($iet_host) )
	{
		logger("err> unable to find host with volume group: $target_vg");
		return undef;
	}
	logger("rest> using iet host: $iet_host");
	
	#
	# determine the source volume's device size
	#
	my $source_dev = "/dev/zvol/dsk/$source_volume";
	logger("rest> source device: $source_dev");
	if ( ! -e "$source_dev" )
	{
		logger("err> unable to find source_dev: $source_dev");
		exit(1);
	}

	#
	# ddless reqires the target to be >= than source), lvm can handle about
	# any init, we are choosing MB
	#
	my $source_size_bytes = sun_disk_size($source_dev);
	logger("rest> source size: $source_size_bytes bytes");
	my $source_size_mb = ceil($source_size_bytes / 1024 / 1024);
	logger("rest> source size: $source_size_mb MB");
	
	my $initiator_name = iscsiadm_initiator_name();
	logger("rest> initiator name: $initiator_name");

	my $vg_lv = "$target_vg/$target_volume";
	logger("rest> vg_lv: $vg_lv");
	
	my %lun_meta = iet_lv_restore($iet_host, $vg_lv, $source_size_mb, $initiator_name);
	if ( !defined(%lun_meta) )
	{
		logger_err("err> no meta data received");
		return undef;
	}
	
	#
	# get target IQN
	#
	my $target_iqn = $lun_meta{"target"};
	logger ("rest> target IQN: $target_iqn");

	#
	# connect via SUN iscsi initiator to the equallogic unit
	#
	my %san_conn = (
		'san_ip' => $cfg{'iet_cluster'}{$iet_host}{'discovery_ip'},
		'san_ip_port' => '3260',
		'san_target' => $target_iqn,
	);

	my $ccode = iscsiadm_connect(\%san_conn);
	if ( !defined($ccode) )
	{
		logger("err> unable to connect with equallogic via iscsi initiator");
		return undef;
	}
	
	#
	# iet has one LUN per target - we do not need to enumerate
	# the LUNs then - not technically correct, however, we only
	# implement a single LUN per target(!)
	#
	my $target_dev = $san_conn{'luns'}{'0'}{'device'};
	if ( !defined($target_dev) )
	{
		logger("err> unable to find LUN 0 device name");
		return undef;
	}

	#
	# with iet, the target device appears to just end with the slice
	# delimiter and no slice number (strange). The problem comes from:
	#
	# scsiadm list target -S
	#
	# OS Device Name: /dev/rdsk/c0t494554000000000044534B2D313030353038323230373431d0s
	#
	logger("rest> source device $target_dev");
	$target_dev =~ s/s$/s2/;
	logger("rest> source device $target_dev (named adjusted)");

	#
	# The target_dev must change the slice 2 reference from the device
	# to partition 0. In SUN tradition slice 2 was an overlapping slice
	# that referred to the entire disk.
	#
	$target_dev =~ s/s2$/p0/;
	logger("rest> target device $target_dev");

	my %device_info = sun_disk_info($target_dev);
	logger("rest> target device serial number ".$device_info{'inquiry-serial-no'});
	logger("rest> target device size $device_info{'disk_size_bytes'} bytes ($device_info{'disk_size_gb'} GB)");

	#
	# 2tb check (pre, post step not needed during restore)
	#
	my $disk_bug = disk_read_2TB_no_part_bug_before_ddless($target_dev);
	if ( !defined($disk_bug) )
	{
		logger_err("err> device size problem encountered!!!");
		return undef;
	}

	#
	# restore data using ddless
	#
	my $cmd = "$cfg{'ddless'} -v -s $source_dev -t $target_dev -c $checksum_file -w 4";
	logger("rest> $cmd");
	system($cmd);
	if ( $? != 0 )
	{
		logger_err("ddless_err> cmd $cmd");
		logger_err("ddless_err> child exit value ".($?>>8));
		return undef;
	}	

	#
	# disconnect iscsi initiator session from iet device
	#
	$ccode = iscsiadm_disconnect(\%san_conn);
	if ( !defined($ccode) )
	{
		logger_err("err> unable to disconnect from iet using iscsi initiator");
		return undef;
	}
	
	#
	# remove iet snapshot and target
	#
	if ( !defined(iet_lv_restoreremove($iet_host, $vg_lv, $initiator_name)) )
	{
		logger_err("err> unable to remove snapshot");
		return undef;
	}

}

