#
# SUN 2Tb volume work around for ddless
#
package Sun2Tb;

use strict;
use POSIX;
require Exporter;

use common;
use config;
use Data::Dumper;

use SunDiskInfo;
use File::Basename;

our @ISA = qw(Exporter);
our @EXPORT = qw(
        disk_read_2TB_no_part_bug_before_ddless
        disk_read_2TB_no_part_bug_after_ddless
	);

#
# This module handles disks that have no partitions and are larger than
# 2TB in size. We have found that disks smaller than 2TB are not impacted
# by this issue. If you have a disk greater than 2TB and you seek to the
# end of the disk and then obtain the current location you might encounter
# the result of 0. This happens if the disk has NO partition. 
#
# If you use the systems' prtconf or DKIOCGMEDIAINFO data structure you
# can obtain the actual disk size irregardless of the partition status.
# Ok, so use that instead of fseek - not so fast! We found that if you
# proceed under those circumstances, that your system will read like this:
# MODE SENSE, READ, MODE SENSE, READ, ... and each mode sense response
# contains the size of the disk - the performance is about 30MB/s - no good.
#
# So, to fix this, on Solaris, we have to use format and 'label' the
# partition. The label is actually an EFI partition label and once that is
# in place, you can read the disk normally and at full speed.
#
# The EFI label is within the first 64KB of the disk and the last 64KB of
# the disk. We will preserve those regions and restore them after ddless is
# complete.
#
# Steffen Plotner, 2010
#

#-------------------------------------------------------------------------------
# run before ddless (the $device is the source device)
#-------------------------------------------------------------------------------
sub disk_read_2TB_no_part_bug_before_ddless
{
	my($device, $volume_name) = @_;

	#
	# obtain kernel level disk info
	#
	my %device_info = sun_disk_info($device);
	my $device_blocksize = $device_info{'device-blksize'};
	my $device_nblocks = $device_info{'device-nblocks'};
	my $device_size_bytes = $device_info{'disk_size_bytes'};
	
	logger("d2> device: $device (source)");
	logger("d2> device_blocksize: $device_blocksize");
	logger("d2> device_nblocks: $device_nblocks");
	logger("d2> device_size: $device_size_bytes bytes");
	
	#
	# obtain disk info via the libc
	#
	my $file_size_bytes = sun_disk_size($device);
	logger("d2> file_size: $file_size_bytes bytes");
	
	if ( $file_size_bytes == 0 )
	{
		#
		# fall through until we return 2
		#
		logger("d2> fixup required...");
	}
	elsif ( $file_size_bytes != $device_size_bytes )
	{
		#
		# we have a major problem here that needs investigation!
		#
		logger("d2_err> $device has different disk size vs file size!!!!");
		return undef;
	}
	else
	{
		#
		# ok
		#
		logger("d2> complete, ok");
		return 1;
	}
	
	#
	# read the first 64kb and save it (this will be slow!)
	#
	my $save_blocks = (64 * 1024 ) / $device_blocksize;
	logger("d2> save blocks: $save_blocks (from start and end of disk)");

	if ( defined($volume_name) )
	{
		my $cmd = "$cfg{'dd'} if=$device of=/tmp/d2_${volume_name}_start ".
			"bs=$device_blocksize count=$save_blocks";
		logger("d2> $cmd");
		system($cmd);
		if ( $? != 0 )
		{
			logger("d2_err> dd failed");
			return undef;
		}
	}
	
	#
	# read the last 64kb and save it (again slow!)
	#
	my $skip_blocks = $device_nblocks - $save_blocks;
	
	if ( defined($volume_name) )
	{
		my $cmd = "$cfg{'dd'} if=$device of=/tmp/d2_${volume_name}_end ".
			"bs=$device_blocksize count=$save_blocks skip=$skip_blocks";
		logger("d2> $cmd");
		system($cmd);
		if ( $? != 0 )
		{
			logger("d2_err> dd failed");
			return undef;
		}
	}
	
	#
	# use the format command to label the disk
	#
	# convert device path to just a device name that format uses
	#
	# /dev/rdsk/c0t494554000000000044534B2D313030363231323330323041d0p0
	# c0t494554000000000044534B2D313030363231323330323041d0
	#
	# The EFI partition label is by spec placed at the beginning and end
	# of the device
	#
	my $device_format = $device;
	$device_format =~ s/p0$//;
	$device_format = basename($device_format);
	logger("d2> device_format: $device_format");	

	my $cmd = "format -f $cfg{'etc_directory'}/format_label ".
		"-d $device_format";
	logger("d2> $cmd");
	system($cmd);
	if ( $? != 0 )
	{
		logger("d2_err> format failed");
		return undef;
	}
	
	if ( defined($volume_name) )
	{
		#
		# restore the last 64kb at this point, since it is possible to
		# read the disk without the backup partition table in place, also
		# it saves on the ddless signature update headache coming up
		#
		my $cmd = "$cfg{'dd'} if=/tmp/d2_${volume_name}_end of=$device ".
			"bs=$device_blocksize count=$save_blocks seek=$skip_blocks";
		logger("d2> $cmd");
		system($cmd);
		if ( $? != 0 )
		{
			logger("d2_err> dd failed");
			return undef;
		}
		
		#
		# cleanup
		#
		unlink("/tmp/d2_${volume_name}_end");
	}

	return 2;	
}

#-------------------------------------------------------------------------------
# run after ddless (the $device is the target device)
#-------------------------------------------------------------------------------
sub disk_read_2TB_no_part_bug_after_ddless
{
	my($device, $checksum_file, $volume_name) = @_;

	logger("d2> device: $device (target)");

	#
	# compute ddless signature of first 64kb
	#
	my $cmd = "$cfg{'ddless'} -v -s /tmp/d2_${volume_name}_start ".
		"-c /tmp/d2_${volume_name}_start.ddless";
	logger("d2> $cmd");
	system($cmd);
	if ( $? != 0 )
	{
		logger("d2_err> ddless post processing failed");
		return undef;
	}
	
	#
	# copy the first 64kb onto the target device (i.e. restore
	# that set of data, it currently contains the EFI partition
	# info)
	#
	my $cmd = "$cfg{'dd'} if=/tmp/d2_${volume_name}_start of=$device";
	logger("d2> $cmd");
	system($cmd);
	if ( $? != 0 )
	{
		logger("d2_err> ddless post processing failed");
		return undef;
	}
	
	#
	# copy the previously re-computed ddless signature onto the
	# signature file of the target device (without truncating the
	# signature file).
	#
	my $cmd = "$cfg{'dd'} if=/tmp/d2_${volume_name}_start.ddless ".
		"of=$checksum_file conv=notrunc,nocreat";
	logger("d2> $cmd");
	system($cmd);
	if ( $? != 0 )
	{
		logger("d2_err> ddless post processing failed");
		return undef;
	}
	
	#
	# cleanup
	#
	unlink("/tmp/d2_${volume_name}_start");
	unlink("/tmp/d2_${volume_name}_start.ddless");
	return 1;
}

