#
# SUN iSCSI Initiator wrapper for the command line tool iscsiadm (Steffen Plotner)
#
package SunIscsiInitiator;

use strict;
require Exporter;

use common;
use Data::Dumper;

our @ISA = qw(Exporter);
our @EXPORT = qw(
	iscsiadm_connect
	iscsiadm_disconnect
	iscsiadm_initiator_name
        );

my $resource = "/tmp/santools-iscsiadm";

#-------------------------------------------------------------------------------
# HIGH LEVEL API
#-------------------------------------------------------------------------------

#-------------------------------------------------------------------------------
# san block connect: this is a high level connect function that calls
# the appropriate low level function in this module to connect to a
# iscsi SAN target and returns the found LUNs from that target
#
# The san_conn hash is expected to look like:
#my %san_conn = (
#	'san_ip' => '148.85.141.20',
#	'san_ip_port' => '3260',
#	'san_target' => "iqn.2001-05.com.equallogic:0-8a0906-9b58d1b01-a64001c831f48ffd-lv-abc-sun-test",
#);
# and by the time the function completes, the resulting LUNs are also part of
# the above san_conn, i.e. pass the block as a hash reference \%.
#-------------------------------------------------------------------------------
sub iscsiadm_connect
{
	my($san_conn) = @_;
	
	#
	# lock (added locking since multiple processes make a mess when modifying
	# the initiator via iscsiadm).
	#
	my $lock = resource_lock($resource);
	if ( !defined($lock) )
	{
		logger_err("iscsiadm_err> unable to obtain resource lock $resource");
		return undef;
	}

	if ( ! defined(iscsiadm_init()) )
	{
		logger_err("iscsiadm_err> unable to init with iscsiadm");
		return undef;
	}

	my %targets = iscsiadm_sendtargets($san_conn);
	if ( !defined(%targets))
	{
		logger_err("iscsiadm_err> unable to add discovery_address or find target");
		return undef;
	}

	if ( !exists($targets{ $$san_conn{'san_target'} }) )
	{
		logger_err("iscsiadm_err> unable to find target $$san_conn{'san_target'} in target list");
		return undef;
	}
	logger("iscsiadm> found san_target=$$san_conn{'san_target'}");

	#
	# the equallogic device only supports LUN 0 per target
	#
	my %san_luns = iscsiadm_bind_target($san_conn);
	$$san_conn{'luns'} = \%san_luns;

	#
	# unlock
	#
	resource_unlock($lock, $resource);
	
	return 1;
}

#-------------------------------------------------------------------------------
# san block disconnect
#-------------------------------------------------------------------------------
sub iscsiadm_disconnect
{
	my($san_conn) = @_;

	#
	# lock
	#
	my $lock = resource_lock($resource);
	if ( !defined($lock) )
	{
		logger_err("iscsiadm_err> unable to obtain resource lock $resource");
		return undef;
	}

	#
	# unbind
	#
	my $ccode = iscsiadm_unbind_target($san_conn);
	if ( !defined($ccode) )
	{
		logger_err("iscsiadm_err> unable to unbind from target");
		return undef;
	}
	delete $$san_conn{'luns'};

	#
	# unlock
	#
	resource_unlock($lock, $resource);

	return 1;
}

#-------------------------------------------------------------------------------
# return current iscsi initiator name
#-------------------------------------------------------------------------------
sub iscsiadm_initiator_name
{
	my $cmd = "iscsiadm list initiator-node";
	logger("iscsiadm> $cmd");
	
	my $out = `$cmd`;
	my $node_name;
	foreach my $line (split /\n/, $out)
	{
		if ( $line =~ /Initiator node name: (\S+)/i )
		{
			$node_name = $1;
		}
	}
	return $node_name;
}

#-------------------------------------------------------------------------------
# LOW LEVEL API
#-------------------------------------------------------------------------------

#-------------------------------------------------------------------------------
# iscsiadm init
#
# initialize the iscsi initiator's discovery setting to static-config
# we have found that dynamic discovery is not needed since you can perform
# the iscsi sendtargets command using iscsiadm to determine the potentially
# new targets.
#-------------------------------------------------------------------------------
sub iscsiadm_init
{
	#Discovery:
	#        Static: enabled		<- this has to be enabled
	#        Send Targets: disabled
	#        iSNS: disabled
	my $cmd = "iscsiadm list discovery";
	logger("iscsiadm> $cmd");

	my $out = `$cmd`;
	foreach my $line (split /\n/, $out)
	{
		if ( $line =~ /Static: disabled/i )
		{
			logger("iscsiadm> static discovery is disabled, enabling it");
			$cmd = "iscsiadm modify discovery --static enable";
			logger("iscsiadm> $cmd");
			system($cmd);
			if ( $? != 0 )
			{
				logger_err("iscsiadm_err> cmd $cmd");
				logger_err("iscsiadm_err> child exit value ".($?>>8));
				return undef;
			}
		}
	}
	return 1;
}

#-------------------------------------------------------------------------------
# Add a specific IP to the discovery_address list, adding them multiple
# times is fine, the underlying iscsiadm tool does the right thing.
# Then issue a sendtargets command and return a list of targets.
#-------------------------------------------------------------------------------
sub iscsiadm_sendtargets
{
	my ($san_conn) = @_;

	my $ip = $$san_conn{'san_ip'};
	my $port = $$san_conn{'san_ip_port'};
	
	#
	# the key is the target name and the value is a hash consisting
	# of the target address and TPGT (target portal group tag)
	#
	my %targets;

	#
	# ip discovery-address command
	#
	my $cmd = "iscsiadm add discovery-address $ip:$port";
	logger("iscsiadm> $cmd");
	system($cmd);
	if ( $? != 0 )
	{
		logger_err("iscsiadm_err> cmd $cmd");
		logger_err("iscsiadm_err> child exit value ".($?>>8));
		return undef;
	}
	
	#
	# sendtargets to get a list of targets
	#
	#Discovery Address: 148.85.141.20:3260
    #    Target name: iqn.2001-05.com.equallogic:0-8a0906-9b58d1b01-a64001c831f48ffd-lv-abc-sun-test
    #            Target address:   148.85.141.20:3260, 0
    #
	$cmd = "iscsiadm list discovery-address -v $ip:$port";
	logger("iscsiadm> $cmd");
	my $out = `$cmd`;
	my $target_name;
	foreach my $line (split /\n/, $out)
	{
		if ( $line =~ /Target name: (.*)/i )
		{
			$target_name = $1;
		}
		if ( $line =~ /Target address:\s+(.*), (\d+)/i )
		{
			$targets{$target_name}{'target_address'} = $1;
			$targets{$target_name}{'target_portal_group_tag'} = $2;
		}
	}
	
	return (%targets);
}

#-------------------------------------------------------------------------------
# Given a target, bind (static-config) with it and return a hash of 
# associated LUNs. Each LUN has a Vendor, Product, Device Name
#-------------------------------------------------------------------------------
sub iscsiadm_bind_target
{
	my($san_conn) = @_;
	
	#
	# once we are bound to the target, enum the LUNs and return the device
	# names, keyed on the LUN id
	#
	my %luns;
	
	#
	# Static Configuration Target: iqn.1990....
	#
	my $cmd = "iscsiadm list static-config";
	logger("iscsiadm> $cmd");
	my @lines = `$cmd`;
	my $target = "$$san_conn{'san_target'},$$san_conn{'san_ip'}:$$san_conn{'san_ip_port'}";
	logger("iscsiadm> search for $target");
	if ( grep(/Static Configuration Target: $target$/i, @lines) == 0)
	{
		logger("iscsiadm> not connected, attempting to bind now...");
		$cmd = "iscsiadm add static-config $target";
		logger("iscsiadm> $cmd");
		system($cmd);
		if ( $? != 0 )
		{
			logger_err("iscsiadm_err> cmd $cmd");
			logger_err("iscsiadm_err> child exit value ".($?>>8));
			return undef;
		}

		#
		# there can be delays before the LUN is actually available and functional
		#
		logger("iscsiadm> waiting 15 sec");
		sleep(15);
	}
	
	#
	# iscsiadm list target (include device info)
	#
	$cmd = "iscsiadm list target --scsi-target $target";
	logger("iscsiadm> $cmd");
	my @lines = `$cmd`;
	my $lun;
	foreach my $line (@lines)
	{
		#
		# LUN: nn
		#
		if ( $line =~ /LUN: (\d+)/ )
		{
			$lun = $1;
		}
		
		#
		# Vendor, Product, Device Name
		#
		if ( defined($lun) && $line =~ /Vendor:\s+(.*?)\s+/ )
		{
			$luns{$lun}{'vendor'} = $1;
		}
		if ( defined($lun) && $line =~ /Product:\s+(.*?)\s+/ )
		{
			$luns{$lun}{'product'} = $1;
		}
		if ( defined($lun) && $line =~ /Device Name:\s+(.*?)\s+/ )
		{
			$luns{$lun}{'device'} = $1;
		}
	}
	return (%luns);
}

#-------------------------------------------------------------------------------
# Given a target, unbind (static-config) from it
#-------------------------------------------------------------------------------
sub iscsiadm_unbind_target
{
	my($san_conn) = @_;

	my $cmd = "iscsiadm list static-config";
	logger("iscsiadm> $cmd");
	my @lines = `$cmd`;
	my $target = "$$san_conn{'san_target'},$$san_conn{'san_ip'}:$$san_conn{'san_ip_port'}";
	logger("iscsiadm> search for $target");
	if ( grep(/Static Configuration Target: $target$/i, @lines) == 1)
	{
		logger("iscsiadm> connected, attempting to unbind in 30 sec...");
		sleep(30);
		$cmd = "iscsiadm remove static-config $target";
		logger("iscsiadm> $cmd");
		system($cmd);
		if ( $? != 0 )
		{
			logger_err("iscsiadm_err> cmd $cmd");
			logger_err("iscsiadm_err> child exit value ".($?>>8));
			return undef;
		}
		logger("iscsiadm> pausing for 15 sec to let things settle...");
		sleep(15);
	}
}

