#
# SUN ZFS wrapper for the command line tool zpool/zfs (Steffen Plotner)
#
package SunZfs;

use strict;
use POSIX;
require Exporter;

use common;
use Data::Dumper;

use SunDiskInfo;

our @ISA = qw(Exporter);
our @EXPORT = qw(
	zfs_create
	zfs_create_snapshot
	zfs_destroy_snapshot
	zfs_zvol_timestamp
	zfs_list_pools
	zfs_list_fsvols
	zfs_list_zvols
	zfs_list_by_prop
	zfs_props
	zfs_snapshots
	zfs_set
        );

#-------------------------------------------------------------------------------
# local/remote API wrapper
#-------------------------------------------------------------------------------
sub zfs_wrapper
{
	my ($zfs_host, $zfs_cmd) = @_;
	
	if ( !defined($zfs_host) )
	{
		$zfs_host = "localhost";
	}
	if ( $zfs_host eq "localhost" )
	{
		return "$zfs_cmd";
	}
	return "ssh $zfs_host $zfs_cmd";
}

#-------------------------------------------------------------------------------
# Given a zfs pool check that we have a zvol defined for lv_name
# if not create it, otherwise check the size of destination volume
# change it if needed (it must be larger)
#
# No remote host support, since we utilize fopen on the disk device to
# determine size... maybe this can be changed to use the remote's disk_size.plx
# code instead?
#-------------------------------------------------------------------------------
sub zfs_create
{
	my ($zfs_pool, $zvol, $zvolsize_bytes, $blocksize_bytes, $zfs_opts) = @_;
	
	#
	# verify pool name
	#
	my $cmd = "zpool list -H -o name $zfs_pool";
	logger("zfs> $cmd");
	system($cmd);
	if ( $? != 0 )
	{
		logger_err("zfs_err> cmd $cmd");
		logger_err("zfs_err> child exit value ".($?>>8));
		return undef;
	}
	
	#
	# check if the volume exists in pool
	#
	$cmd = "zfs list -H -o name $zfs_pool/$zvol";
	logger("zfs> $cmd");
	system($cmd);
	if ( $? != 0 )
	{
		#
		# creating the zvol requires a block size of 16k as this matches
		# what ddless uses - otherwise the actual snap storage consumption
		# will not line up.
		#
		# on solaris 10, the volume size must be a multiple of the blocksize
		# later versions do not require this
		#
		my $block_count = 0;
		if ($zvolsize_bytes % $blocksize_bytes)
		{
			$block_count = 1;
		}
		$block_count += int($zvolsize_bytes / $blocksize_bytes);
		my $vol_size_bytes = $block_count * $blocksize_bytes;
		logger("zfs> rounded to nearest blocksize $vol_size_bytes");
		
		$cmd = "zfs create -b $blocksize_bytes -V $vol_size_bytes $zfs_opts $zfs_pool/$zvol";
		logger("zfs> $cmd");
		system($cmd);
		if ( $? != 0 )
		{
			logger_err("zfs_err> cmd $cmd");
			logger_err("zfs_err> child exit value ".($?>>8));
			return undef;
		}
	}
	
	#
	# check if the volume needs to be enlarged?
	#
	# zfs list -H -o volsize tank/lv-name outputs the result like 5G (including
	# units)
	#
	my $destination_bytes = sun_disk_size("/dev/zvol/dsk/$zfs_pool/$zvol");
	logger("zfs> volume $zvol size $destination_bytes bytes");
	
	if ( $zvolsize_bytes > $destination_bytes )
	{
		logger("zfs> the destination device $zvol must be grown");

		$cmd = "zfs set volsize=$zvolsize_bytes $zfs_pool/$zvol";
		logger("zfs> $cmd");
		system($cmd);
		if ( $? != 0 )
		{
			logger_err("zfs_err> cmd $cmd");
			logger_err("zfs_err> child exit value ".($?>>8));
			return undef;
		}
		#
		# volume got resized
		#
		return 2;
	}
	return 1;
}

#-------------------------------------------------------------------------------
# create a zfs snapshot given pool/volume and snapshot name
#-------------------------------------------------------------------------------
sub zfs_create_snapshot
{
	my ($zfs_host, $zfs_pool, $zvol, $snap_name) = @_;

	my $snapshot_name = "$zfs_pool/$zvol\@$snap_name";
	my 	$cmd = zfs_wrapper($zfs_host,"zfs snapshot $snapshot_name");
	logger("zfs> $cmd");
	system($cmd);
	if ( $? != 0 )
	{
		logger_err("zfs_err> cmd $cmd");
		logger_err("zfs_err> child exit value ".($?>>8));
		return undef;
	}
	zfs_set($zfs_host, $snapshot_name, 'edu.amherst:snap_time', time());
	return $snapshot_name;
}

#-------------------------------------------------------------------------------
# destroy a zfs snapshot
#-------------------------------------------------------------------------------
sub zfs_destroy_snapshot
{
	my ($zfs_host, $snapshot_name) = @_;

	if ( $snapshot_name !~ /\@/ )
	{
		logger_err("zfs_err> this appears not to be a snapshot: $snapshot_name");
		return undef;
	}
	
	my 	$cmd = zfs_wrapper($zfs_host,"zfs destroy $snapshot_name");
	logger("zfs> $cmd");
	system($cmd);
	if ( $? != 0 )
	{
		logger_err("zfs_err> cmd $cmd");
		logger_err("zfs_err> child exit value ".($?>>8));
		return undef;
	}
	return 1;
}

#-------------------------------------------------------------------------------
# make zvolume snapshot name
#-------------------------------------------------------------------------------
sub zfs_zvol_timestamp
{
	my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst)=localtime(time);
	$year=$year+1900;
	$mon++;
	my $event_date = sprintf("%04d-%02d-%02d_%02d:%02d:%02d",$year,$mon,$mday,$hour,$min,$sec);
	return "$event_date";
}

#-------------------------------------------------------------------------------
# prepare a list of zfs zpools
#-------------------------------------------------------------------------------
sub zfs_list_zpools
{
	my ($zfs_host) = @_;
	my @list;
	my $cmd = zfs_wrapper($zfs_host,"zpool list -o name -H");
	logger("zfs> $cmd");
	my $out = `$cmd`;
	return (split /\n/,$out);
}

#-------------------------------------------------------------------------------
# prepare a list of zfs file systems and volumes
#-------------------------------------------------------------------------------
sub zfs_list_fsvols
{
	my ($zfs_host) = @_;
	my @list;
	my $cmd = zfs_wrapper($zfs_host,"zfs list -H -t volume,filesystem -o name");
	logger("zfs> $cmd");
	my $out = `$cmd`;
	return (split /\n/,$out);
}

#-------------------------------------------------------------------------------
# prepare a list of zfs volumes
#-------------------------------------------------------------------------------
sub zfs_list_zvols
{
	my ($zfs_host) = @_;
	my @list;
	my $cmd = zfs_wrapper($zfs_host,"zfs list -H -t volume -o name");
	logger("zfs> $cmd");
	my $out = `$cmd`;
	return (split /\n/,$out);
}
#-------------------------------------------------------------------------------
# query the zfs list command for a given property
#-------------------------------------------------------------------------------
sub zfs_list_by_prop
{
	my ($zfs_host, $zfs_fld, $zfs_val) = @_;
	
	my @list;
	my $cmd = zfs_wrapper($zfs_host,"zfs list -H -t filesystem,volume -o name,$zfs_fld");
	logger("zfs> $cmd");
	my $out = `$cmd`;
	foreach my $line (split /\n/,$out)
	{
		#logger("line=$line");
		if ( defined($zfs_fld) && 
			defined($zfs_val) 
			&& $line =~ /^(.*?)\t(.*?)$/ )
		{
			my $zfs_name = $1;
			my $zfs_prop_val = $2;
			if ( $zfs_prop_val eq $zfs_val )
			{
				push @list, $zfs_name;
			}
		}
		else
		{
			push @list, $line
		}
	}
	return @list;
}

#-------------------------------------------------------------------------------
# given a zfs name, obtain the properties as a hash reference
#-------------------------------------------------------------------------------
sub zfs_props
{
	my ($zfs_host, $zfs_name, $no_units) = @_;
	
	my $no_units_args = "";
	if ( defined($no_units) && $no_units == 1 )
	{
		$no_units_args = "-p";
	}
	
	my %props;
	my $cmd = zfs_wrapper($zfs_host,"zfs get $no_units_args -H all $zfs_name");
	#logger("zfs> $cmd");
	my $out = `$cmd`;
	foreach my $line (split /\n/,$out)
	{
		if ( $line =~ /^(.*?)\t(.*?)\t(.*?)\t(.*?)$/ )
		{
			my $name = $1;
			my $property = $2;
			my $value = $3;
			
			$props{$property} = $value;
		}
	}
	
	return \%props;
}

#-------------------------------------------------------------------------------
# get the snapshot names given a zfs filesystem/volume
#-------------------------------------------------------------------------------
sub zfs_snapshots
{
	my ($zfs_host, $zfs_name) = @_;
	
	my @list;
	my $cmd = zfs_wrapper($zfs_host,"zfs list -o name -H -r -t snapshot $zfs_name");
	logger("zfs> $cmd");
	my $out = `$cmd`;
	foreach my $line (split /\n/,$out)
	{
		#logger("line=$line");
		push @list, $line
	}
	return @list;
}

#-------------------------------------------------------------------------------
# set zfs property
#-------------------------------------------------------------------------------
sub zfs_set
{
	my ($zfs_host, $zfs_name, $prop, $value) = @_;

	my $cmd = zfs_wrapper($zfs_host,"zfs set $prop=\"$value\" $zfs_name");
	logger("zfs> $cmd");
	system($cmd);
	if ( $? != 0 )
	{
		logger_err("zfs_err> cmd $cmd");
		logger_err("zfs_err> child exit value ".($?>>8));
		return undef;
	}
}

