#
# EQL Wrapper
#
# Steffen Plotner, 2005-2010
#
package eqlutil;
use strict;
require Exporter;

use common;
use Data::Dumper;

#
# some of the function below return arrays instead of keyed hashes, because
# some of the equallogic 'show' functions returns lists that do not have real
# unique keys, example is the access control list
#

our @ISA = qw(Exporter);
our @EXPORT = qw(
	eql_connect
	eql_disconnect
	eql_list_dump
	eql_members
	eql_volumes
	eql_volume_lookup
	eql_grpparams_show
	eql_pool_show
	eql_pool_lookup
	eql_pool_select_show
	eql_volume_select_snapshot_create_now
	eql_volume_select_snapshot_delete
	eql_volume_select_snapshot_show
	eql_snapshot_lookup
	eql_volume_select_show
	eql_volume_select_show_access
	eql_volume_select_show_members
	eql_volume_access_lookup
	eql_volume_select_access_create
	eql_volume_select_access_delete
	eql_volume_select_snapshot_select_operation
	eql_volume_select_snapshot_select_show
	eql_volume_set_snap_reserve
	eql_volume_set_snap_warn
	eql_volume_select_clone
	eql_volume_select_schedule_show
	eql_volume_schedule_lookup
	eql_volume_schedule_manage
	eql_volume_create
        );

my @pool_fields = (
	"Name",
	"Description",
	"Default",
	"TotalVolumes",
	"VolumesOnline",
	"VolumesInUse",
	"TotalSnapshots",
	"SnapshotsOnline",
	"SnapshotsInUse",
	"TotalMembers",
	"MembersInUse",
	"TotalCapacity",
	"VolumeSpace",
	"SnapReservedSpace",
	"SnapReservedSpaceInUse",
	"ReplicationReservedSpace",
	"DelegatedSpace",
	"DelegatedSpaceInUse",
	"FreeSpace",
);

my @volume_fields = (
	"Name",
	"Size",
	"VolReserve",
	"VolReserveInUse",
	"ReplReserveInUse",
	"iSCSI Alias",
	"iSCSI Name",
	"ActualMembers",
	"Snap-Warn",
	"Snap-Depletion",
	"Description",
	"Snap-Reserve",
	"Snap-Reserve-Avail",
	"Permission",
	"DesiredStatus",
	"Status",
	"Connections",
	"Snapshots",
	"Bind",
	"Type",
	"ReplicationReserveSpace",
	"Replicas",
	"ReplicationPartner",
	"Pool",
	"Transmitted-Data",
	"Received-Data",
	"Pref-Raid-Policy",
	"Pref-Raid-Policy-Status",
);

my @schedule_fields =
(
	"Name",
	"Type",
	"AccessType",
	"Start-Date",
	"End-Date",
	"Start-Time",
	"End-Time",
	"Frequency",
	"Repeat-Interval",
	"Max-Keep",
	"NextCreation",
	"Desired Status",
	"Status",
);

my @snap_fields =
(
	"Name",
	"Permission",
	"DesiredStatus",
	"Status",
	"OriginalVolumeSize",
	"CreationTime",
	"Schedule",
	"Replicate",
	"iSCSI Alias",
	"iSCSI Name",
	"Description",
	"Connections",
	"SnapsetName",
);

#-------------------------------------------------------------------------------
# dump an array of hashes, allows one to specify which keys in the hashes to include
#-------------------------------------------------------------------------------
sub array_hash_dump
{
	my ($array_hash, $prefix, @fields) = @_;
	
	foreach my $entry (@$array_hash)
	{
		my $line = "$prefix ";
		
		foreach my $field (@fields)
		{
			$line .= "($field=".$$entry{$field}.")";
		}
		
		logger($line);
	}
}

#-------------------------------------------------------------------------------
# lookup data in an array of hashes
#-------------------------------------------------------------------------------
sub array_hash_lookup
{
	my ($array_hash, $field, $value) = @_;

	if ( ! defined($array_hash) )
	{
		return undef;
	}
	
	foreach my $entry (@$array_hash)
	{
		#logger("DEBUG> ".$$entry{$field}."=".$value);
		if ( $$entry{$field} eq $value )
		{
			return $entry;
		}
	}
	return undef;
}

#-------------------------------------------------------------------------------
#
# connect with equallogic device
#
#-------------------------------------------------------------------------------
sub eql_connect
{
	my ($ip, $user, $pass) = @_;

	my $eql = EqlScript->new();
	if ( !$eql )
	{
		logger("eql_err> unable to create EqlScript object");
		return 0;
	}
	
	#
	# we are having issues with the volume show command taking
	# too long
	#
	$eql->timeout(60);

	logger("eql> connecting to $ip");
	if ( !$eql->open(Group => $ip) )
	{
		logger("eql_err> unable to connect with $ip");
		return 0;
	}

	logger("eql> logging on as $user");
	if ( !$eql->login(Account => $user, Password => $pass) )
	{
		logger("eql_err> unable to authenticate");
		return 0;
	}
	logger("eql> connected!");

	my $cmd = "cli-settings events off";
	logger("eql_cmd> $cmd");
	$eql->cmd($cmd);

	$cmd = "stty hardwrap off";
	logger("eql_cmd> $cmd");
	$eql->cmd($cmd);

	$cmd = "stty columns 255";
	logger("eql_cmd> $cmd");
	$eql->cmd($cmd);

	return $eql;
}

#-------------------------------------------------------------------------------
# disconnect from eql
#-------------------------------------------------------------------------------
sub eql_disconnect
{
	my ($eql) = @_;
	
	my $cmd = "logout";
	logger("eql_cmd> $cmd");
	$eql->cmd($cmd);
}

#-------------------------------------------------------------------------------
# get an eql list 
#
# get a command line based list, append each record to an array, use a hash value
# structure to identify the field->value associations (for each record!)
# array of hash.
#
# lists are output as follows by the eql (including the wrapping - arg!)
#
#Name            Size       SnapShots Status         Permission Connections 
#--------------- ---------- --------- -------------- ---------- ----------- 
#lv-mail5-boot-c 12GB       0         online         read-write 1           
#lv-mail5-exchlo 60GB       0         online         read-write 1           
#  g-e                                                                      
#lv-mail5-exchmd 120GB      0         online         read-write 1           
#  b-d                                                                      
#
#-------------------------------------------------------------------------------
sub eql_get_list
{
	my ($eql, $cmd) = @_;
	
	logger("eql_cmd> $cmd");
	my @lines = $eql->cmd($cmd);
	if ( $eql->err )
	{
		logger("eql_err> ".$eql->errmsg);
		return undef;
	}

	# require a header, followed by dashes
	if ( scalar(@lines) < 2 )
	{
		logger("eql_err> less than 2 lines returned by list command: $cmd");
		
		foreach my $line (@lines)
		{
			logger("eql_err> line returned: $line");
		}
		return undef;
	}
	
	# volume select __ show members has a column name with spaces!
	$lines[0] =~ s/Member Name/Member_Name/;
	
	# parse the first line's fields into a hash (fieldname->length of field)
	my @field_names = split(/\s* +\s*/,shift(@lines));
	my $field_name_count = scalar(@field_names);
	
	# parse the second line's dashes, to determine each field's length
	my @field_lengths = split(/\s* +\s*/,shift(@lines));
	my $field_length_count = scalar(@field_lengths);
	
	if ( $field_name_count != $field_length_count )
	{
		logger("eql_err> the number of colum headings $field_name_count does ".
			"not match the number of column heading ".
			"dividers (dashes) $field_length_count");
		logger("eql_err> fields: ".join(',',@field_names));
		logger("eql_err> lengths: ".join(',',@field_lengths));
		return undef;
	}
	
	# determine the field lengths based on the number of dashes
	for ( my $i=0; $i < scalar(@field_lengths); $i++)
	{
		$field_lengths[$i] = length($field_lengths[$i]);
	}
		
	#print Dumper \@field_names;
	#print Dumper \@field_lengths;
	#print Dumper \@lines;	
	
	# process data lines
	my @data;
	my $idx;
	my $blank_lines = 0;
	while( my $line = shift(@lines) )
	{
		#
		# skip blank lines, if we find 2 exit
		#
		chomp($line);
		chomp($line);
		#logger("!$line!");
		if ( $line eq "" )
		{
			#logger("BL> $blank_lines");	
			$blank_lines++;
			if ( $blank_lines >= 2 )
			{
				last;
			}
			next;
		}
		
		$idx = scalar(@data);
		
		# check if line has 2 blank spaces in position 0, 1, if so it's a 
		# wrapper and the data has to go to the previous row
		#logger("$line");
		if ( $line =~ /^  / )
		{
			$idx--;
		}

		my $pos = 0;
		for ( my $i=0; $i < scalar(@field_names); $i++)
		{
			my $val = substr($line, $pos, $field_lengths[$i]);
			$val =~ s/ +$//;
			$val =~ s/^ +//;
			$data[$idx]{$field_names[$i]} .= $val;
			$pos += $field_lengths[$i] + 1;
		}
	}
	
	return \@data;
}

#-------------------------------------------------------------------------------
# given a list produced by eql_get_list, dump it based on the field names requested
#-------------------------------------------------------------------------------
sub eql_list_dump
{
	my ($list, $prefix, @fields) = @_;
	
	array_hash_dump($list, $prefix, @fields);
}

#-------------------------------------------------------------------------------
# eql get report
# 
# exec eql logic command, given an array of fields, pattern match them, 
# return a hash (key=field, value=value of field)
#
#-------------------------------------------------------------------------------
sub eql_get_report
{
	my ($eql, $cmd, @fields) = @_;
	
	logger("eql_cmd> $cmd");
	my @lines = $eql->cmd($cmd);
	if ( $eql->err )
	{
		logger("eql_err> ".$eql->errmsg);
		return undef;
	}
	
	my %data;
	
	foreach my $field (@fields)
	{
		#logger("field=$field");
		foreach my $line (@lines)
		{
			if ( $line =~ /^$field: (.*)/ )
			{
				#logger("fieldvalue=$1");
				$data{$field} = $1;
			}
		}
	}
	
	return \%data;
}

#-------------------------------------------------------------------------------
# get a list of members (the detail data is not included, just the list)
#-------------------------------------------------------------------------------
sub eql_members
{
	my ($eql) = @_;
	
	return eql_get_list($eql,"member show");
}

#-------------------------------------------------------------------------------
# get a list of volumes and their associated attributes
#
# list all volumes and return an array of hashes (field->value)
#-------------------------------------------------------------------------------
sub eql_volumes
{
	my ($eql) = @_;
	
	my $volumes = eql_get_list($eql,"volume show");
	foreach my $volume (@$volumes)
	{
		my $rep = eql_get_report($eql, "volume select $$volume{'Name'} show", @volume_fields);

		foreach my $field ( keys %{$rep} )
		{
			$$volume{$field} = $$rep{$field};
		}
	}
	return $volumes
}


#-------------------------------------------------------------------------------
# volume select show
#
# given a volume, get detailed volume information
#-------------------------------------------------------------------------------
sub eql_volume_select_show
{
	my ($eql,$volume_name) = @_;
	
	my $rep = eql_get_report($eql, "volume select $volume_name show", @volume_fields);

	my %volume;
	foreach my $field ( keys %{$rep} )
	{
		$volume{$field} = $$rep{$field};
	}
	return \%volume;
}

#-------------------------------------------------------------------------------
# search for volume
#-------------------------------------------------------------------------------
sub eql_volume_lookup
{
	my ($volumes, $volume_name) = @_;
	return array_hash_lookup($volumes,'Name',$volume_name);
}

#-------------------------------------------------------------------------------
# get group parameters
#
# returns a reference to a hash containing group parameters (key->value)
#-------------------------------------------------------------------------------
sub eql_grpparams_show
{
	my ($eql) = @_;

	my $rep = eql_get_report($eql, "grpparam show", ("Name", "Group-Ipaddress"));
	return $rep;
}

#-------------------------------------------------------------------------------
# get pool parameters
#
# returns a reference to a hash containing group parameters (key->value)
#-------------------------------------------------------------------------------
sub eql_pool_show
{
	my ($eql) = @_;

	return eql_get_list($eql,"pool show");
}

#-------------------------------------------------------------------------------
# search for pool
#-------------------------------------------------------------------------------
sub eql_pool_lookup
{
	my ($pool, $pool_name) = @_;
	return array_hash_lookup($pool,'Name',$pool_name);
}

#-------------------------------------------------------------------------------
# search for pool
#-------------------------------------------------------------------------------
sub eql_pool_select_show
{
	my ($eql, $pool_name) = @_;

	my $rep = eql_get_report($eql, "pool select $pool_name show", 
		@pool_fields);
	return $rep;
}

#-------------------------------------------------------------------------------
# create snapshot and return name based on volume name
#-------------------------------------------------------------------------------
sub eql_volume_select_snapshot_create_now
{
	my ($eql, $volume_name) = @_;
	
	my $cmd = "volume select $volume_name snapshot create-now";
	logger("eql_cmd> $cmd");
	my @lines = $eql->cmd($cmd);
	if ( $eql->err )
	{
		logger("eql_err> ".$eql->errmsg);
		return undef;
	}

	#
	# Snapshot creation succeeded.
	# Snapshot name is test-2008-01-08-21:48:15.26389
	#
	if ( scalar(@lines) >= 2 && 
		$lines[0] =~ /snapshot creation succeeded/i &&
		$lines[1] =~ /snapshot name is (.*)/i )
	{
		return $1;
	}
	return undef;
}

#-------------------------------------------------------------------------------
# volume snapshots delete
#-------------------------------------------------------------------------------
sub eql_volume_select_snapshot_delete
{
	my ($eql,$volume_name, $snapshot_name) = @_;
	
	my $cmd = "volume select $volume_name snapshot delete $snapshot_name";
	logger("eql_cmd> $cmd");
	my @lines = $eql->cmd($cmd);
	if ( $eql->err )
	{
		logger("eql_err> ".$eql->errmsg);
		return undef;
	}
	logger("eql_out> ".join(', ',@lines));
}

#-------------------------------------------------------------------------------
# volume snapshots show
#
# list all snapshots of a volume and return an array of hashes (field->value)
#-------------------------------------------------------------------------------
sub eql_volume_select_snapshot_show
{
	my ($eql, $volume_name) = @_;
	
	return eql_get_list($eql,"volume select $volume_name snapshot show");
}

#-------------------------------------------------------------------------------
# snapshot lookup
#
# given a snapshot array reference, produced by eql_volume_select_snapshot_show, lookup 
# a snapshot by name
#-------------------------------------------------------------------------------
sub eql_snapshot_lookup
{
	my ($snapshots, $snapshot_name) = @_;

	return array_hash_lookup($snapshots,'Name',$snapshot_name);
}

#-------------------------------------------------------------------------------
# volume access control list show
#
# list acls based on volume and return an array of hashes (field->value)
#-------------------------------------------------------------------------------
sub eql_volume_select_show_access
{
	my ($eql, $volume_name) = @_;
	
	return eql_get_list($eql,"volume select $volume_name show access");
}

#-------------------------------------------------------------------------------
# volume members list show
#
# list members of a volume and return an array of hashes (field->value)
#-------------------------------------------------------------------------------
sub eql_volume_select_show_members
{
	my ($eql, $volume_name) = @_;
	
	return eql_get_list($eql,"volume select $volume_name show members");
}

#-------------------------------------------------------------------------------
# volume access lookup
#
# given a acl array reference, produced by eql_volume_select_show_access, 
# lookup a acl entry by field name
#-------------------------------------------------------------------------------
sub eql_volume_access_lookup
{
	my ($acls, $field, $value) = @_;

	return array_hash_lookup($acls,$field,$value);
}

#-------------------------------------------------------------------------------
# volume access create
#-------------------------------------------------------------------------------
sub eql_volume_select_access_create
{
	my ($eql, $volume_name, $access) = @_;
	
	my $cmd = "volume select $volume_name access create $access";
	logger("eql_cmd> $cmd");
	my @lines = $eql->cmd($cmd);
	if ( $eql->err )
	{
		logger("eql_err> ".$eql->errmsg);
		return undef;
	}
	logger("eql_out> ".join(', ',@lines));
	return 1;
}

#-------------------------------------------------------------------------------
# volume access delete
#-------------------------------------------------------------------------------
sub eql_volume_select_access_delete
{
	my ($eql, $volume_name, $acl_id) = @_;
	
	my $cmd = "volume select $volume_name access delete $acl_id";
	logger("eql_cmd> $cmd");
	my @lines = $eql->cmd($cmd);
	if ( $eql->err )
	{
		logger("eql_err> ".$eql->errmsg);
		return undef;
	}
	logger("eql_out> ".join(', ',@lines));
	return 1;
}

#-------------------------------------------------------------------------------
# volume snapshot operations
#-------------------------------------------------------------------------------
sub eql_volume_select_snapshot_select_operation
{
	my ($eql, $volume_name, $snapshot_name, $operation) = @_;
	
	my $cmd = "volume select $volume_name snapshot select $snapshot_name $operation";
	logger("eql_cmd> $cmd");
	my @lines = $eql->cmd($cmd);
	if ( $eql->err )
	{
		logger("eql_err> ".$eql->errmsg);
		return undef;
	}
	logger("eql_out> ".join(', ',@lines));
	return 1;
}

#-------------------------------------------------------------------------------
# volume snapshot information
#-------------------------------------------------------------------------------
sub eql_volume_select_snapshot_select_show
{
	my ($eql, $volume_name, $snapshot_name) = @_;

	my $rep = eql_get_report($eql, "volume select $volume_name snapshot select $snapshot_name show", 
		@snap_fields);
	return $rep;
}

#-------------------------------------------------------------------------------
# volume change snap-reserve
#-------------------------------------------------------------------------------
sub eql_volume_set_snap_reserve
{
	my ($eql,$volume_name, $snap_reserve_percent) = @_;
	
	my $cmd = "volume select $volume_name snap-reserve ${snap_reserve_percent}%";
	logger("eql_cmd> $cmd");
	my @lines = $eql->cmd($cmd);
	if ( $eql->err )
	{
		logger("eql_err> ".$eql->errmsg);
		return undef;
	}
	logger("eql_out> ".join(', ',@lines));
}

#-------------------------------------------------------------------------------
# volume change snap-warn
#-------------------------------------------------------------------------------
sub eql_volume_set_snap_warn
{
	my ($eql,$volume_name, $snap_warn_percent) = @_;
	
	my $cmd = "volume select $volume_name snap-warn ${snap_warn_percent}%";
	logger("eql_cmd> $cmd");
	my @lines = $eql->cmd($cmd);
	if ( $eql->err )
	{
		logger("eql_err> ".$eql->errmsg);
		return undef;
	}
	logger("eql_out> ".join(', ',@lines));
}

#-------------------------------------------------------------------------------
# volume clone
#-------------------------------------------------------------------------------
sub eql_volume_select_clone
{
	my ($eql, $volume_name, $volume_name_new, $volume_desc_new) = @_;

	my $cmd = "volume select $volume_name clone $volume_name_new description \"$volume_desc_new\"";
	logger("eql_cmd> $cmd");
	my @lines = $eql->cmd($cmd);
	if ( $eql->err )
	{
		logger("eql_err> ".$eql->errmsg);
		return undef;
	}
	logger("eql_out> ".join(', ',@lines));
}

#-------------------------------------------------------------------------------
# volume schedules (array) (BUG: eql also emits replication schedule, we don't
# support that and neither do we support multiple lists from a single command)
#
# list all schedules of a volume and return an array of hashes (field->value)
#-------------------------------------------------------------------------------
sub eql_volume_select_schedule_show
{
	my ($eql, $volume_name) = @_;
	
	my $schedules = eql_get_list($eql,"volume select $volume_name schedule show");
	foreach my $schedule (@$schedules)
	{
		#
		# get the schedule's detail
		#
		my $name = $$schedule{'Name'};
		my $rep = eql_get_report($eql, "volume select $volume_name schedule show $name", 
			@schedule_fields);

		foreach my $field ( keys %{$rep} )
		{
			$$schedule{$field} = $$rep{$field};
		}
	}
	return $schedules
}

#-------------------------------------------------------------------------------
# volume schedule lookup
#
# given a schedule array reference, produced by eql_volume_schedule_show, 
# lookup a acl entry by field name
#-------------------------------------------------------------------------------
sub eql_volume_schedule_lookup
{
	my ($schedules, $value) = @_;

	return array_hash_lookup($schedules,'Name',$value);
}

#-------------------------------------------------------------------------------
# volume schedule control (expects create/select (update)
#-------------------------------------------------------------------------------
sub eql_volume_schedule_manage
{
	my ($eql, $volume_name, $schedule_definition) = @_;
	
	my $cmd = "volume select $volume_name $schedule_definition";
	logger("eql_cmd> $cmd");
	my @lines = $eql->cmd($cmd);
	if ( $eql->err )
	{
		logger("eql_err> ".$eql->errmsg);
		return undef;
	}
	logger("eql_out> ".join(', ',@lines));
	return 1;
}

#-------------------------------------------------------------------------------
# volume create
#-------------------------------------------------------------------------------
sub eql_volume_create
{
	my ($eql, $volume_name, $volume_size_mb) = @_;
	
	my $cmd = "volume create $volume_name ${volume_size_mb}MB";
	logger("eql_cmd> $cmd");
	my @lines = $eql->cmd($cmd);
	if ( $eql->err )
	{
		logger("eql_err> ".$eql->errmsg);
		return undef;
	}
	logger("eql_out> ".join(', ',@lines));
	return 1;
}

