#
# IET Wrapper
#
# 2010, Steffen Plotner
#
package ietutil;
use strict;
require Exporter;

use common;
use config;
use Data::Dumper;
use Sys::Hostname;

our @ISA = qw(Exporter);
our @EXPORT = qw(
	iet_vgcheck
	iet_volumes
	iet_lv_snap
	iet_lv_snapremove
	iet_lv_dirty
	iet_lv_restore
	iet_lv_restoreremove
	iet_ddmap_fetch
        );

my $host = hostname;

#-------------------------------------------------------------------------------
# ssh argument wrapper to support custom pub/priv keys
#-------------------------------------------------------------------------------
sub iet_ssh_args
{
	return "-i $cfg{'etc_directory'}/id_rsa_iscsi_target";
}

#-------------------------------------------------------------------------------
# check if volume group is mounted on host
#-------------------------------------------------------------------------------
sub iet_vgcheck
{
	my ($iet_host, $vg) = @_;
	
	my $cmd = "ssh ".iet_ssh_args()." $iet_host ".
		"/etc/iscsi-target/zfs_tools.sh vgcheck \"$vg\"";
	logger("iet> $cmd");
	my $line = `$cmd`;
	chomp($line);
	logger("iet> result: $line");

	if ( $line eq $vg )
	{
		return 1;
	}
	return 0;
}

#-------------------------------------------------------------------------------
# enumerate iet hosts and gather a complete list of logical volumes and
# volume groups, the result is a hash with keys consisting of vg/lv. This
# function filters out logical volumes that only pertain to this host.
#-------------------------------------------------------------------------------
sub iet_volumes
{
	my %volumes;
	foreach my $iet_host (keys %{$cfg{'iet_cluster'}})
	{
		if ( !$cfg{'iet_cluster'}{$iet_host}{'enabled'} )
		{
			next;
		}

		logger("iet> iet_host: $iet_host");
		my $cmd = "ssh ".iet_ssh_args()." $iet_host ".
			"/etc/iscsi-target/zfs_tools.sh lvlist";
		logger("iet> $cmd");
		foreach my $key (split(/\n/, `$cmd`))
		{
			my ($vg, $lv) = split(/\//, $key);
			
			#
			# lookup the volume group mapping based on the current host
			#
			if ( exists($cfg{'iet_luncopy_map'}{$host}{'pool_map'}{$vg}) )
			{
				logger("iet> vg_lv: $key (this host)");
				$volumes{$key}{'dest_host'} = $host;
				$volumes{$key}{'iet_vg'} = $vg;
				$volumes{$key}{'iet_lv'} = $lv;
				$volumes{$key}{'iet_host'} = $iet_host;
				$volumes{$key}{'dest_pool'} = $cfg{'iet_luncopy_map'}{$host}{'pool_map'}{$vg}
			}
			else
			{
				logger("iet> vg_lv: $key (other host)");
			}
		}
	}
	return %volumes;
}

#-------------------------------------------------------------------------------
# iet create a logical volume snapshot and prepare the ddmap for it
#-------------------------------------------------------------------------------
sub iet_lv_snap
{
	my ($iet_host, $vg_lv, $initiator_name, $integrity_check) = @_;
	
	my %snap_meta;
	
	my $cmd = "ssh ".iet_ssh_args()." $iet_host ".
		"/etc/iscsi-target/zfs_tools.sh lvsnap $vg_lv $initiator_name $integrity_check";
	logger("iet> $cmd");
	foreach my $line (split(/\n/, `$cmd`))
	{
		logger("iet> $line");
		if ( $line =~ /DATA> target: (\S+)/ )
		{
			$snap_meta{'target'} = $1;
		}
		elsif ( $line =~ /DATA> ddmap_state: (\S+)/ )
		{
			$snap_meta{'ddmap_state'} = $1;
		}
		elsif ( $line =~ /DATA> ddmap_file: (\S+)/ )
		{
			$snap_meta{'ddmap_file'} = $1;
		}
		elsif ( $line =~ /DATA> ddless_md5: (\S+)/ )
		{
			$snap_meta{'ddless_md5'} = $1;
		}
	}
	return %snap_meta;
}

#-------------------------------------------------------------------------------
# iet snapshot remove
#-------------------------------------------------------------------------------
sub iet_lv_snapremove
{
	my ($iet_host, $vg_lv, $initiator_name) = @_;
		
	my $cmd = "ssh ".iet_ssh_args()." $iet_host ".
		"/etc/iscsi-target/zfs_tools.sh lvsnapremove $vg_lv $initiator_name";
	logger("iet> $cmd");
	system($cmd);
	if ( $? != 0 )
	{
		logger_err("iet_err> cmd $cmd");
		logger_err("iet_err> child exit value ".($?>>8));
		return undef;
	}
	return 0;
}

#-------------------------------------------------------------------------------
# iet flag volume as dirty on integrity errors
#-------------------------------------------------------------------------------
sub iet_lv_dirty
{
	my ($iet_host, $vg_lv) = @_;
	
	my $cmd = "ssh ".iet_ssh_args()." $iet_host ".
		"/etc/iscsi-target/zfs_tools.sh lvdirty $vg_lv";
	logger("iet> $cmd");
	system($cmd);
	if ( $? != 0 )
	{
		logger_err("iet_err> cmd $cmd");
		logger_err("iet_err> child exit value ".($?>>8));
		return undef;
	}
}

#-------------------------------------------------------------------------------
# iet restore logical volume, prepare lun and expose target
#-------------------------------------------------------------------------------
sub iet_lv_restore
{
	my ($iet_host, $vg_lv, $lun_size, $initiator_name) = @_;
	
	my %lun_meta;
	
	my $cmd = "ssh ".iet_ssh_args()." $iet_host ".
		"/etc/iscsi-target/zfs_tools.sh lvrestore \"$vg_lv\" \"$lun_size\" \"$initiator_name\"";
	logger("iet> $cmd");
	foreach my $line (split(/\n/, `$cmd`))
	{
		logger("iet> $line");
		if ( $line =~ /DATA> target: (\S+)/ )
		{
			$lun_meta{'target'} = $1;
		}
	}
	return %lun_meta;
}

#-------------------------------------------------------------------------------
# iet restore logical volume, remove target
#-------------------------------------------------------------------------------
sub iet_lv_restoreremove
{
	my ($iet_host, $vg_lv, $initiator_name) = @_;
		
	my $cmd = "ssh ".iet_ssh_args()." $iet_host ".
		"/etc/iscsi-target/zfs_tools.sh lvrestoreremove $vg_lv $initiator_name";
	logger("iet> $cmd");
	system($cmd);
	if ( $? != 0 )
	{
		logger_err("iet_err> cmd $cmd");
		logger_err("iet_err> child exit value ".($?>>8));
		return undef;
	}
	return 0;
}

#-------------------------------------------------------------------------------
# scp remote ddmap to a local ddmap
#-------------------------------------------------------------------------------
sub iet_ddmap_fetch
{
	my ($iet_host, $ddmap_remote, $ddmap_local) = @_;
	
	my $cmd = "scp ".iet_ssh_args()." $iet_host:$ddmap_remote $ddmap_local";
	logger("iet> $cmd");
	system($cmd);
	if ( $? != 0 )
	{
		logger_err("iet_err> cmd $cmd");
		logger_err("iet_err> child exit value ".($?>>8));
		return undef;
	}
}


