#!/usr/bin/perl

# ZFS Snapshot Checker, by Steffen Plotner, 2010

use strict;
use Data::Dumper;
use Getopt::Std;
use File::Copy;
use Sys::Hostname;
use Digest::MD5;

use FindBin;
use lib "$FindBin::Bin/lib";
use lib "$FindBin::Bin/etc";

# amherst provided
use common;
use config;
use SunZfs;

$cfg{'verbose'} = 1;

#
# handle command line options
#
my %opts;
getopts('hs:do',\%opts);
if ( defined($opts{'h'}) )
{
	print <<EOH;

$0 [-h] -s zpool/zname\@snapshot | -d | -o

This code uses ddless to compute the ddless checksum of a given snapshot and
compares it to the ddless checksums in the repository. The -s allows you to
check a single snapshot adhoc. The -d is meant to be invoked by cron and
verifies snapshots continuously.

-h	help
-s	snapshot name to check
-d	daily check of $cfg{'snapcheck'}{'check_count'} snapshots
-o	delete orphaned ddless signature files (i.e. snapshot has been deleted and
	ddless signature is left behind - the snaprotate code does not delete the
	ddless signature files.)
EOH
	exit(0);
}

if ( defined($opts{'s'}) )
{
	my $snapcheck = $opts{'s'};
	snapshot_check($snapcheck);
}
elsif ( defined($opts{'d'}) )
{
	snapshot_checkjob();
}
elsif ( defined($opts{'o'}) )
{
	snapshot_ddless_orphans();
}
exit(0);

#-------------------------------------------------------------------------------
# validate a single snapshot
#-------------------------------------------------------------------------------
sub snapshot_check
{
	my($snapname) = @_;

	logger("check> $snapname");
	
	#
	# snapshot names must have the name structure:
	#
	# vg_satabeast1_vol0/g-elogic0_lv-mail5-boot-c_bk@2009-09-03_00:04:30
	#
	my $zpool;
	my $zvol;
	my $zdate;
	if ( $snapname =~ /(\S+)\/(\S+)_bk\@(\S+)/ )
	{
		$zpool = $1;
		$zvol = $2;
		$zdate = $3;
	}
	logger("check> zpool: $zpool zvol: $zvol zdate: $zdate");
	if ( ! (defined($zpool) && defined($zvol) && defined($zdate)) )
	{
		logger("check_err> unable to parse snapshot name: $snapname");
		return undef;
	}
	
	#
	# given the parsed snapshot info, calculate the ddless snapshot checksum
	# file location
	#
	my $ddless_hist = "/$zpool/ddless/$zvol.history/${zvol}_${zdate}.ddless";
	logger("check> ddlesss_hist: $ddless_hist");
	if ( ! -e $ddless_hist )
	{
		logger("check_err> unable to find ddless_hist: $ddless_hist");
		return undef;
	}
	
	#
	# compute digest of historical checksum
	#
	if ( !open(FH, "<$ddless_hist") )
	{
		logger("check_err> unable to open ddless_hist: $ddless_hist");
		return undef;
	}
	binmode(FH);
	my $ddless_hist_md5 = Digest::MD5->new->addfile(*FH)->hexdigest();
	close(FH);
	logger("check> ddless_hist_md5: $ddless_hist_md5");
	
	#
	# compute ddless checksum of snapshot using ddless
	#
	my $ddless_check = "/tmp/snapcheck.$zpool.$zvol.$zdate.ddless";
	if ( -e $ddless_check )
	{
		unlink($ddless_check);
	}

	my $zdev = "/dev/zvol/rdsk/$snapname";
	my $cmd = "$cfg{'ddless'} -s $zdev -c $ddless_check -w $cfg{'snapcheck'}{'ddless_workers'}";
	logger("check> spawn $cmd");
	system($cmd);
	if ( $? != 0 )
	{
		logger_err("check_err> cmd $cmd");
		logger_err("check_err> child exit value ".($?>>8));
		return undef;
	}
	
	#
	# compute digest of current checksum
	#
	if ( !open(FH, "<$ddless_check") )
	{
		logger("check_err> unable to open ddless_check: $ddless_check");
		return undef;
	}
	binmode(FH);
	my $ddless_check_md5 = Digest::MD5->new->addfile(*FH)->hexdigest();
	close(FH);
	logger("check> ddless_check_md5: $ddless_check_md5");
	
	#
	# compare
	#
	if ( $ddless_hist_md5 eq $ddless_check_md5 )
	{
		logger("check_RESULT> PASS $snapname");
	}
	else
	{
		logger("check_RESULT> FAIL $snapname (next: ddless_maint)");
	}
	unlink($ddless_check);
	logger();
}

#-------------------------------------------------------------------------------
# snapshot check fire event - schedule volumes for checking and validate n
# snapshots of scheduled volumes
#-------------------------------------------------------------------------------
sub snapshot_checkjob
{
	logger("check> daily run");
	
	my $state_file = "$cfg{'state_directory'}/snapcheck.dat";

	my @zvols;
	
	#
	# state: first time, create a new file
	#
	if ( ! -e "$state_file" )
	{
		logger("check> no existing state file, creating new one");
		if ( !open(FH,">$state_file") )
		{
			logger("check_err> unable to create file $state_file: $?");
			return undef;
		}
		close(FH);
	}
	
	#
	# read existing state
	#
	if ( !open(FH, "<$state_file") )
	{	
		logger("check_err> unable to open file $state_file: $?");
		return undef;
	}
	while(<FH>)
	{
		my $line = $_;
		chomp($line);
		if ( $line ne "" )
		{
			push @zvols, $line;
		}
	}
	close(FH);
	
	#
	# how many zvols do we have?
	#
	my $zvol_count = scalar(@zvols);
	logger("check> read $zvol_count volumes");
	
	#
	# if there are no zvols, then populate the state using all zvols from system
	#
	if ( $zvol_count == 0 )
	{
		@zvols = zfs_list_zvols(undef);
	}
	
	#
	# worker loop - we are going to validate at least n snapshots
	#
	my $did_work = 0;
	my $max_work = $cfg{'snapcheck'}{'check_count_per_run'};
	logger("check> max_work: $max_work");
	while ( $did_work < $max_work)
	{
		#
		# extract the top volume name (and remove that from the list)
		#
		my $zvol = pop @zvols;
		if ( !defined($zvol) )
		{
			#
			# do not reload the volume list - even so it might make sense
			# you might have an infinite loop if all volumes have old snapshots
			# - so just cron schedule us for the next day when we do nothing
			# again in this case!
			#
			logger("check> no more volumes on the list");
			last;
		}
		logger("check> zvol: $zvol");
	
		#
		# obtain a list of its snapshots
		#
		my @snapshots = zfs_snapshots(undef, $zvol);
		if ( scalar(@snapshots) == 0 )
		{
			logger("check> no snapshots found, skipping zvol");
			next;
		}
		my $snapshot = pop @snapshots;
		logger("check> snapshot: $snapshot");
		
		#
		# get the properties and look for: edu.amherst:snap_time
		#
		my $props = zfs_props(undef, $snapshot);
		my $snap_time = $$props{'edu.amherst:snap_time'};
		if ( !defined($snap_time) )
		{
			logger("check> snapshot has no snap_time, skipping zvol");
			next;
		}
		logger("check> snap_time: $snap_time");
		
		#
		# time diff (ignore last snapshots of a volume, if the snapshot is old)
		#
		my $age_seconds = time() - $snap_time;
		my $age_days = $age_seconds / (24 * 60 * 60);
		logger("check> age: $age_seconds seconds, $age_days days");
		
		if ( $age_days > 7 )
		{
			logger("check> snapshot too old, skipping zvol");
			next;
		}
		
		#
		# process the ddless signatures of the last snapshot
		#
		logger("check> processing $snapshot ...");
		
		snapshot_check($snapshot);
		
		$did_work++;
		logger("check> did_work: $did_work");
	}
	
	#
	# write out the state
	#
	if ( !open(FH, ">$state_file") )
	{
		logger("check_err> unable to open file $state_file: $?");
		return undef;
	}
	foreach my $zvol (@zvols)
	{
		print FH "$zvol\n";
	}
	close(FH);
}

#-------------------------------------------------------------------------------
# delete ddless checksum files for which there is no related snapshot
#-------------------------------------------------------------------------------
sub snapshot_ddless_orphans
{
	logger("check> looking for ddless signatures without related snapshot");

	#
	# first get all zvols
	#
	my @zvols_bk = zfs_list_zvols(undef);
	foreach my $zvol_bk (@zvols_bk)
	{
		#
		# parse the zpool/zvol_bk (the _bk will not be part of the ddless
		# signature history, because the zvol_bk was not used to create the
		# ddless signature, the original iscsi volume from the source was
		# used to create the ddless signature.
		#
		my ($zpool, $zvol);
		logger("check> looking at zpoolzvol: $zvol_bk");
		if ( $zvol_bk =~ /(\S+)\/(\S+)_bk/ )
		{
			$zpool = $1;
			$zvol = $2;
		}
		
		logger("check> zpool: $zpool zvol: $zvol");
		if ( ! (defined($zpool) && defined($zvol)) )
		{
			logger("check> unable to parse zvol_bk: $zvol_bk, skipping it");
			next;
		}

		#
		# track the ddless signature files in a hash with the key being
		# the name of the original zfs snapshot
		#
		my %data;
		
		#
		# prepare a glob filter for the ddless signatures
		#
		# file: /vg_satabeast0_vol1/ddless/g-elogic0_lv-cms-copy-vol00.history/
		#		g-elogic0_lv-cms-copy-vol00_2009-07-06_06:04:56.ddless
		#
		# key: vg_satabeast0_vol1/g-elogic0_lv-cms-copy-vol00_bk@2009-07-05_06:03:52
		#
		my $ddless_hist_glob = "/$zpool/ddless/$zvol.history/*.ddless";
		logger("check> ddlesss_hist: $ddless_hist_glob");
		my @ddless_hist = glob($ddless_hist_glob);
		#print Dumper \@ddless_hist;
		logger("check> found ".scalar(@ddless_hist)." ddless files");
		if ( scalar(@ddless_hist) == 0 )
		{
			logger("check> no ddless files found: $zvol_bk, skipping it");
			next;
		}
		
		foreach my $ddless_file (@ddless_hist)
		{
			if ( $ddless_file =~ /.*\/(\S+)_(\d\d\d\d-\d\d-\d\d)_(\d\d:\d\d:\d\d)\.ddless$/)
			{
				my $lv = $1;
				my $date = $2;
				my $time = $3;
				my $snapshot = "$zpool/${lv}_bk\@${date}_${time}";
				$data{$snapshot}{'ddless'} = $ddless_file;
				$data{$snapshot}{'keep'} = 0;
			}
			else
			{
				logger("check_err> pattern mismatch: $ddless_file");
			}
		}
		#print Dumper \%data;
		
		#
		# get the actual snapshots of this volume
		#
		my @snapshots = zfs_snapshots(undef, $zvol_bk);
		#print Dumper \@snapshots;
		
		foreach my $snapshot (@snapshots)
		{
			logger("check> keeping ddless file for snapshot: $snapshot");
			if ( !defined($data{$snapshot}) )
			{
				logger("check_err> we have a snapshot and no associated ddless file");
				logger("check_err> snapshot: $snapshot");
				next;
			}
			$data{$snapshot}{'keep'} = 1;
		}
		#print Dumper \%data;

		#
		# enum over the data hash and remove ddless that are unused
		#
		foreach my $snapshot (keys %data)
		{
			if ( $data{$snapshot}{'keep'} == 0 )
			{
				my $ddless_file = $data{$snapshot}{'ddless'};
				logger("check> removing ddless_file: $ddless_file");
				unlink($ddless_file);
			}
		}
	}
}

