#!/usr/bin/perl

#
# snaprotate: given a zfs volume or filesystem, rotate the dependend
# snapshots according to a schedule. The schedule can include monthly,
# weekly and daily components indicating the retention time requirements.
#
use strict;
use Data::Dumper;
use Getopt::Std;

use FindBin;
use lib "$FindBin::Bin/lib";
use lib "$FindBin::Bin/etc";

# amherst provided
use common;
use config;
use SunZfs;

$cfg{'verbose'} = 1;

my %rule_day_factor_table = (
	'daily' => 1,
	'weekly' => 7,
	'monthly' => 30,
);

#
# handle command line options
#
my %opts;
getopts('has:n',\%opts);
if ( defined($opts{'h'}) )
{
	print <<EOH;
$0 [-h] -s zfs_object | -a [-n]

-h	help
-a	process all zfs volumes and filesystems, only choose those that have the
	custom property edu.amherst:snaprotate_schedule defined
-s	process a specific zfs volume or filesystems, the custom property
	edu.amherst:snaprotate_schedule must be defined
-n	simulate, no changes are made
EOH
	exit(0);
}

my $simulate = 0;
if ( defined($opts{'n'}) )
{
	$simulate = 1;
}

if ( !defined($opts{'a'}) && !defined($opts{'s'}) )
{
	logger_err("err> must specify either -a or -s, use -h to get help");
	exit(1);
}


if ( defined($opts{'s'}) )
{
	my $zfs_fsvol = $opts{'s'};
	my ($snapshots, $drop_snapshots) = snaprotate($zfs_fsvol, $simulate);
	if ( defined($drop_snapshots) )
	{
		snapdrops($zfs_fsvol, $snapshots, $drop_snapshots, $simulate);
	}
}
else
{
	my $snap_count = 0;
	my $drop_count = 0;
	my @zfs_fsvols = zfs_list_fsvols(undef);
	foreach my $zfs_fsvol (@zfs_fsvols)
	{
		my ($snapshots, $drop_snapshots) = snaprotate($zfs_fsvol, $simulate);
		if ( defined($drop_snapshots) )
		{
			snapdrops($zfs_fsvol, $snapshots, $drop_snapshots, $simulate);
			$drop_count += scalar(@{$drop_snapshots});
			$snap_count += scalar(@{$snapshots});
		}
		logger();
	}
	logger("rot> total number of snapshots dropped: $drop_count out of $snap_count");
}
exit(0);

#-------------------------------------------------------------------------------
# provide an array of snapshots to be dropped
#-------------------------------------------------------------------------------
sub snapdrops
{
	my ($zfs_fsvol, $snapshots, $drop_snapshots, $simulate) = @_;

	logger("rot> preparing to drop ".scalar(@{$drop_snapshots}).
		" of ".scalar(@{$snapshots})." snapshots from $zfs_fsvol");

	foreach my $drop_snapshot (sort @{$drop_snapshots})
	{
		logger("rot> drop $drop_snapshot");

		if ( $simulate )
		{
			logger("rot> -n used, no changes made");
		}
		elsif ( !zfs_destroy_snapshot(undef, $drop_snapshot) )
		{
			logger_err("err> unable to drop snapshot: $drop_snapshot");
		}
	}       
}

#-------------------------------------------------------------------------------
# snap rotate schedule a single zfs volume or file system and return an array
# of snapshots to be dropped
#-------------------------------------------------------------------------------
sub snaprotate
{
	my ($zfs_name, $simulate) = @_;
	
	logger("rot> zfs_name=$zfs_name");

	my $props = zfs_props(undef, $zfs_name);
	if ( !exists($$props{'edu.amherst:snaprotate_schedule'}) )
	{
		logger("rot> custom property edu.amherst:snaprotate_schedule not found");
		return;
	}
	
	my $rotate_schedule = $$props{'edu.amherst:snaprotate_schedule'};
	logger("rot> using schedule $rotate_schedule");
	
	#
	# make sure we have a definition for the found rotate_schedule
	#
	if ( !exists($cfg{'snaprotate_schedules'}{$rotate_schedule}) )
	{
		logger("rot> rotate_schedule: $rotate_schedule found in zfs object, ".
			"but it is not defined in config.pm, skipping over object");
		return;
	}
	
	#
	# clear the rule buckets before entering the loop (the loop system below
	# might not get to all the rule buckets to clear them, it depends on the
	# actual data sets we are processing, so initialize this here and be safe
	#
	foreach my $rule ( @{$cfg{'snaprotate_schedules'}{$rotate_schedule}} )
	{
		$$rule{'bucket'} = ();
	}

	#
	# get the current snapshot instances (process them with the newest snapshot
	# first going back in time, i.e. reverse the list)
	#
	my @snapshots = zfs_snapshots(undef, $zfs_name);
	
	# array index of $cfg{'snaprotate_schedules'}{$rotate_schedule}
	my $rule_index = -1;	

	my $prev_snap_time = time();

	my $rule_day_factor;
	my $rule_age_days = 0;
	my $cumulative_age_days = 0;
	my $cumulative_age_days_max = 0;

	my @drop_snapshots;
	
	foreach my $snapshot (reverse @snapshots)
	{
		#logger("rot> snapshot: $snapshot");

		#
		# get the snapshot timestamp (custom property)
		#
		my $props = zfs_props(undef, $snapshot);
		if ( !exists($$props{'edu.amherst:snap_time'}) )
		{
			logger("rot> custom property edu.amherst:snap_time not found");
			next;
		}
		my $snap_time = $$props{'edu.amherst:snap_time'};
		
		#
		# calculate the age of the snapshot (days), track the previous value
		#
		my $age_delta_days = round(($prev_snap_time - $snap_time) / 86400);
		$prev_snap_time = $snap_time;

		#
		# accumulate the current rule's total age days and the rule age days
		#
		$cumulative_age_days += $age_delta_days;
		$rule_age_days += $age_delta_days;
		
		logger("rot> snap_time: ".snaptime_format($snap_time)." ".
			"age_delta_days: $age_delta_days ".
			"rule_age_days: $rule_age_days ".
			"cumulative_age_days: $cumulative_age_days ".
			"cumulative_age_days_max: $cumulative_age_days_max");

		#logger("DEBUG> age_days: ". (time()-$snap_time)/86400 );
		
		#
		# check if the cumulative age bumps us into the next rule
		#
		my $rule_bump_result = undef;
		
		while ($rule_index == -1 ||
			$cumulative_age_days > $cumulative_age_days_max )
		{
			$rule_index++;
			if ( $rule_index >=
				scalar(@{$cfg{'snaprotate_schedules'}{$rotate_schedule}}) )
			{
				logger("rot> past any defined rules, adding to drop_snapshots array");
				
				#
				# all snaps that are not constraint by rules are past
				# all rules, hence into the bit bucket
				#
				push @drop_snapshots, $snapshot;
				
				$rule_bump_result = 'past_any_rules';
				last;
			}
			my $rule = 
				$cfg{'snaprotate_schedules'}{$rotate_schedule}[$rule_index];
			my $rule_name = $$rule{'name'};
			my $rule_freq = $$rule{'freq'};
			my $rule_duration = $$rule{'duration'};

			$rule_day_factor = $rule_day_factor_table{$rule_freq};
			
			logger("rot> rule index: $rule_index name: $rule_name freq: $rule_freq ".
				"day_factor: $rule_day_factor count: $rule_duration");
			
			#
			# we track this to determine the next rule jump
			#
			my $rules_days = $rule_day_factor * $rule_duration;
			logger("rot> days in this rule: $rules_days");
			
			$cumulative_age_days_max += $rules_days;
			logger("rot> cumulative_age_days_max: $cumulative_age_days_max");
			
			$rule_bump_result = 'reset';
		}
		
		if ( $rule_bump_result eq 'past_any_rules' )
		{
			logger("rot> past any rules, next snapshot...");
			next;
		}
		elsif ( $rule_bump_result eq 'reset' )
		{
			#
			# reset for each new rule, since it needs to be relative to the
			# bucket index of the rule
			#
			$rule_age_days = 0;
			logger("rot> rule_age_days: $rule_age_days");
		}

		#
		# given the cumulative value of the age in days based on the current
		# rule, determine the bucket index. Each rule is assigned an array of
		# buckets. Each bucket can hold one snapshot.
		#
		my $bucket_real = $rule_age_days / $rule_day_factor;
		my $bucket_index = round($bucket_real);
		logger("rot> bucket_index: $bucket_index ($bucket_real)");

		my $rule = 
			$cfg{'snaprotate_schedules'}{$rotate_schedule}[$rule_index];

		#
		# assign a rule a bucket array using bucket_index as the
		# array index and then push the current snapshot into the bucket,
		# i.e. multiple snapshots can land in the same bucket, we have to
		# weed that out after this loop
		#
		push @{$$rule{'bucket'}[$bucket_index]}, $snapshot;
	}
	
	#
	# safe-guard (only if there are actually any snapshots)
	#
	logger("rot> number of snapshots: ".scalar(@snapshots));
	logger("rot> number of snapshots to drop: ".scalar(@drop_snapshots));
	if ( scalar(@snapshots) > 0 && scalar(@snapshots) == scalar(@drop_snapshots) )
	{
		logger_err("rot_err> we are told to drop all snapshots?");
		exit(1);
	}
		
	#print Dumper $cfg{'snaprotate_schedules'};
	#print Dumper \@drop_snapshots;
	
	#
	# iterate over the rules and their buckets, if a bucket contains
	# multiple snapshots, pick the last one and put the others
	# into the drop_snapshots array. There might be values in that
	# array from the above operation already, those are snapshots that
	# have expired all rules
	#
	logger("rot> looking through buckets, only allowing one snapshot per bucket");
	foreach my $rule ( @{$cfg{'snaprotate_schedules'}{$rotate_schedule}} )
	{
		my $rule_name = $$rule{'name'};
		my $rule_freq = $$rule{'freq'};
		my $rule_duration = $$rule{'duration'};
		my $rule_day_factor = $rule_day_factor_table{$rule_freq};
			
		logger("rot> rule name: $rule_name freq: $rule_freq ".
				"day_factor: $rule_day_factor count: $rule_duration");

		my $bucket_index = 0;
		my $bucket_keep_index = 0;
		foreach my $bucket ( @{$$rule{'bucket'}} )
		{
			#print Dumper $bucket;
			if ( ref($bucket) ne 'ARRAY' )
			{
				$bucket_index++;
				next;
			}
			my $count = scalar(@{$bucket});
			
			#
			# if you have 2, the old code would pick the first one to
			# keep and drop the last one, which means the aging never
			# happens - changed the algorithm to keep the oldest one.
			#
			#my $bucket_keep_index = round($count / 2) - 1;
			my $bucket_keep_index = $count -1;
			
			logger("rot> bucket[$bucket_index]: $count item(s), ".
				"keeping snapshot at index $bucket_keep_index");

			#
			# enum over the bucket contents, moving each snapshot to the
			# drop set, however if the index matches the bucket keep index
			# then we keep that one
			#
			for(my $i=0; $i<$count; $i++)
			{
				my $snapshot_name = @{$bucket}[$i];
				
				if ( $i != $bucket_keep_index )
				{
					push @drop_snapshots, $snapshot_name;
					logger("rot> i=$i $snapshot_name (drop)");
				}
				else
				{
					logger("rot> i=$i $snapshot_name (keep)");
					if ( !$simulate)
					{
						zfs_set(undef, $snapshot_name,
							"edu.amherst:snaprotate_rule","$rule_name\[current:$bucket_index max:$rule_duration timeslots\]");
					}
				}
			}
			$bucket_index++;
		}
	}
	
	#print Dumper sort @drop_snapshots;
	
	return (\@snapshots, \@drop_snapshots);
}

#-------------------------------------------------------------------------------
# return current date time string
#-------------------------------------------------------------------------------
sub snaptime_format
{
	my ($time) = @_;
	
	my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst)=localtime($time);
	$year=$year+1900;
	$mon++;
	my $event_date = sprintf("%04d-%02d-%02d %02d:%02d:%02d",$year,$mon,$mday,$hour,$min,$sec);
	return $event_date;
}

#-------------------------------------------------------------------------------
# round both positive and negative numbers
#-------------------------------------------------------------------------------
sub round
{
	my ($number) = @_;
	return int($number + .5 * ($number <=> 0));
}

