#!/usr/bin/perl

#
# Tk ZFS interface to permit the cloning of zfs volumes, exposing
# them to a iscsi target initiator and to remove the clones.
#
# Steffen Plotner, 2009
#
use strict;
use Tk;
use Tk::Tree;
use Tk::Adjuster;
use Tk::Dialog;
use Tk::DialogBox;
use Tk::BrowseEntry;

use Data::Dumper;
use Sys::Hostname;
use English;

use FindBin;
use lib "$FindBin::Bin";
use config;
use tkzfsrepo;
use tkzfsutil;

if ( $OSNAME eq "MSWin32" )
{
	my $dyn_code = "use tkzfswin;";
	eval "$dyn_code";
}
elsif ( $OSNAME eq "linux" )
{
	my $dyn_code = "use tkzfslinux;";
	eval "$dyn_code";
}

#
# discover the current host and its features
#
print "Launching $0 on operating system: $OSNAME with hostname: ".hostname.".\n";

#
# dynamic feature/function dispatch table (they appear to be evaluated late
# as not having the above module loaded, does not cause an error)
#
my %fn_dispatch = (
	'ms_iscsi' =>
		{
			'map' => \&tkzfswin::win32_lun_map_ms_iscsi,
			'lookup' => \&tkzfswin::win32_lun_lookup_ms_iscsi,
			'umap' => \&tkzfswin::win32_lun_umap_ms_iscsi,
		},
	'qla_iscsi' =>
		{
			'map' => \&tkzfswin::win32_lun_map_qla_iscsi,
			'lookup' => \&tkzfswin::win32_lun_lookup_qla_iscsi,
			'umap' => \&tkzfswin::win32_lun_umap_qla_iscsi,
		},
	'open_iscsi' =>
		{
			'map' => \&tkzfslinux::linux_lun_map,
			'lookup' => \&tkzfslinux::linux_lun_lookup,
			'umap' => \&tkzfslinux::linux_lun_umap,
		},
);

#
# do the initial load of the repo data, using cached values
#
zfs_repo_refresh(undef,"fast");

#
# main window
#
my $w_main = MainWindow->new();
$w_main->title("TK ZFS Repository");

#
# top frame (which holds left, middle, right frame)
#
my $w_frametop = $w_main->Frame(
	-relief => 'flat',
	-borderwidth => 2,
	)->pack(
		-side => 'top',
		-fill => 'both',
		-expand => 1,
	);

$w_main->Adjuster(
	)->packAfter(
		$w_frametop,
	);

#
# bottom frame holds the application logger
#
my $w_framebottom = $w_main->Frame(
	-relief => 'flat',
	-borderwidth => 2,
	)->pack(
		-side => 'bottom',
		-fill => 'both',
		-expand => 1,
	);
my $wtxt_logger = $w_framebottom->Scrolled(
	"Text",
	-scrollbars => 'ser',
	-height => 16,
	-width => 60,
	-wrap => 'none',
	)->pack(
		-fill => 'both',
		-expand => 1,
		);

#
# top frame: left frame
#
my $w_frameleft = $w_frametop->Frame(
	-relief => 'flat',
	-borderwidth => 2,
	)->pack(
		-side => 'left',
		-fill => 'y',
	);
	
#
# top frame: left frame: tree
#
my $w_tree = $w_frameleft->Scrolled(
	"Tree",
	-scrollbars => 'ser',
	-itemtype => 'text',
	-separator => '@',
	-height => 30,
	-width => 60,
	-browsecmd => \&zfs_tree_click,
	)->pack(
		-side => 'left',
		-fill => 'both',
		-expand => 1,
	);

$w_frametop->Adjuster(
	)->packAfter(
		$w_frameleft,
		-side => 'left',
	);


#
# top frame: right frame
#
my $w_frameright = $w_frametop->Frame(
	-relief => 'flat',
	-borderwidth => 2,
	)->pack(
		-side => 'left',
		-fill => 'both',
		-expand => 1,
	);

#
# right frame: header
#
my $wfld_zfsheader = $w_frameright->Label(
	-textvariable => \$cfg{'tree_item_display'},
	-relief => 'flat',
	-borderwidth => 1,
	)->pack(
		-side => 'top', 
		);

#
# right frame: info box
#	
my $wtxt_zfsinfo = $w_frameright->Scrolled(
	"Text",
	-scrollbars => 'ser',
	-height => 20,
	-width => 60,
	-wrap => 'none',
	)->pack(
		-fill => 'both',
		-expand => 1,
		);
		
$wtxt_zfsinfo->insert('end',"Currently configured repositories:\n".
	join("\n",keys %{$cfg{'repositories'}}) );

#
# right frame: tools frame
#
my $w_frametools = $w_frameright->Frame(
	-relief => 'flat',
	-borderwidth => 4,
	-height => 10,
	)->pack(
		-fill => 'both',
	);

$w_frametools->Label(
	-text => "Repo: ",
	-relief => 'flat',
	-borderwidth => 1,
	)->pack(
		-side => 'left', 
		);

my $wcmd_refresh = $w_frametools->Button(
	-text => 'Refresh',
	-command => \&cmd_refresh,
	)->pack(-side => 'left');

my $wcmd_clone_create = $w_frametools->Button(
	-text => 'Create Clone',
	-command => \&cmd_clone_create,
	-state => 'disabled',
	)->pack(-side => 'left');

my $wcmd_clone_delete = $w_frametools->Button(
	-text => 'Delete Clone',
	-command => \&cmd_clone_delete,
	-state => 'disabled',
	)->pack(-side => 'left');

my $wcmd_map;
my $wcmd_umap;
if ( $cfg{'os'}{$OSNAME}{'mapclone'} )
{
	$w_frametools->Label(
		-text => " Local: ",
		-relief => 'flat',
		-borderwidth => 1,
		)->pack(
			-side => 'left', 
			);

	$wcmd_map = $w_frametools->Button(
		-text => 'Map Clone',
		-command => \&cmd_map,
		-state => 'disabled',
		)->pack(-side => 'left');

	$wcmd_umap = $w_frametools->Button(
		-text => 'Unmap Clone',
		-command => \&cmd_umap,
		-state => 'disabled',
		)->pack(-side => 'left');
}

#
# tk event loop
#
zfs_tree_make();
#$w_tree->selectionSet("vg_satabeast0_vol1/lv_hct");
#$w_tree->yview("vg_satabeast0_vol1/lv_hct");

#
# initialize the logger
#
logger_init($w_main, $wtxt_logger);
logger("Welcome to the ZFS Repository");
logger("doc> Repo Refresh: refreshes the tree view");
logger("doc> Repo Create Clone: create a clone of an existing snapshot");
logger("doc> Repo Delete Clone: delete a cloned snapshot");
if ( $cfg{'os'}{$OSNAME}{'mapclone'} )
{
	logger("doc> Local Map Clone: attach the clone to this machine and map it with a drive letter");
	logger("doc> Local Unmap Clone: detach the clone from this machine");
}
check_config();
Tk::MainLoop();
exit(0);

#-------------------------------------------------------------------------------
# check configuration
#-------------------------------------------------------------------------------
sub check_config
{
	my $this_host = hostname;
	if ( !exists($cfg{'host'}{$this_host}) )
	{
		logger("startup_err> unable to find hostname: $this_host in configuration module");
	}
	
	#
	# check local vs remote version
	#
	my $local_version = $cfg{'version'};
	
	foreach my $repo (keys %{$cfg{'repositories'}})
	{
		if ( !$cfg{'repositories'}{$repo}{'enabled'} )
		{
			next;
		}
		
		my $repo_version = zfs_repo_version($repo);
		logger("startup> repository: $repo has version: $repo_version");
		if ( $repo_version ne $local_version )
		{
			logger("startup_err> the local version: $local_version does not match the remote repository $repo (version $repo_version)");
		}
	}
}

#-------------------------------------------------------------------------------
# zfs tree make
#-------------------------------------------------------------------------------
sub zfs_tree_make
{
	#
	# enumerate the tree hash
	#
	foreach my $tree_item (sort keys %{$cfg{'tree'}})
	{
		my $tree_val = $cfg{'tree'}{$tree_item}{'display'};

		$w_tree->add("$tree_item",-text=>"$tree_val");

		#
		# close tree branches with snapshots
		#
		if ( defined($cfg{'tree'}{$tree_item}{'properties'}{'_tkzfs_zfs_snapshot'}) )
		{
			my $parent_item = $cfg{'tree'}{$tree_item}{'parent'};
		
			$w_tree->setmode($parent_item,'close');
			$w_tree->close($parent_item);
		}
	}
}

#-------------------------------------------------------------------------------
# zfs tree event
#-------------------------------------------------------------------------------
sub zfs_tree_click
{
	my($tree_item) = @_;

	#
	# prevent double-events from firing
	#
	if ( $cfg{'tree_item'} eq $tree_item )
	{
		return;
	}

	logger("tree> selected: $tree_item");

	my $is_snapshot = defined($cfg{'tree'}{$tree_item}{'properties'}{'_tkzfs_zfs_snapshot'});
	my $is_clone = defined($cfg{'tree'}{$tree_item}{'properties'}{'origin'});
	my $is_share_iscsi = ($cfg{'tree'}{$tree_item}{'properties'}{'shareiscsi'} eq 'on');
	
	#
	# save the currently selected tree item, as it is used by others...
	#
	$cfg{'tree_item'} = $tree_item;
	$cfg{'tree_item_display'} = $cfg{'tree'}{$tree_item}{'properties'}{'_tkzfs_zfs_name'};
	if ( $is_snapshot )
	{
		$cfg{'tree_item_display'} .= '@'.$cfg{'tree'}{$tree_item}{'properties'}{'_tkzfs_zfs_snapshot'};
	}

	#
	# prepare info text
	#	
	my $info;
	foreach my $prop ( sort keys %{$cfg{'tree'}{$tree_item}{'properties'}} )
	{
		$info .= sprintf("%-20s = %s\n",
			$prop, $cfg{'tree'}{$tree_item}{'properties'}{$prop});
	}
	
	#
	# clear info box and fill it in (1.0 means line 1, col 0)
	#
	$wtxt_zfsinfo->delete('1.0','end');
	$wtxt_zfsinfo->insert('end',"$info");

	#
	# allow clone create/delete buttons
	#
	if ( $is_snapshot )
	{
		$wcmd_clone_create->configure(-state => 'normal');
	}
	else
	{
		$wcmd_clone_create->configure(-state => 'disabled');
	}

	if ( $is_clone )
	{
		$wcmd_clone_delete->configure(-state => 'normal');
	}
	else
	{
		$wcmd_clone_delete->configure(-state => 'disabled');
	}

	#
	# clone mapping/unmapping buttons
	#
	if ( $cfg{'os'}{$OSNAME}{'mapclone'} )
	{
		#
		# map button
		#
		if ( $is_clone && $is_share_iscsi )
		{
			$wcmd_map->configure(-state => 'normal');
		}
		else
		{
			$wcmd_map->configure(-state => 'disabled');
		}
		
		#
		# unmap button
		#
		if ( $is_clone && $is_share_iscsi )
		{
			my $this_host = hostname;
			my $iscsi_hba = $cfg{'host'}{$this_host}{'iscsi_hba'};

			#
			# dynamically dispatch the function call
			#
			logger("cmd_map> iscsi_hba: $iscsi_hba");
			my $is_mapped = $fn_dispatch{$iscsi_hba}{'lookup'}($w_main);

			if ( $is_mapped == 0 )
			{
				$wcmd_umap->configure(-state => 'disabled');
			}
			else
			{
				$wcmd_umap->configure(-state => 'normal');
			}
		}
		else
		{
			$wcmd_umap->configure(-state => 'disabled');
		}
			
	}
}

#-------------------------------------------------------------------------------
# refresh data tree
#-------------------------------------------------------------------------------
sub cmd_refresh
{
	my ($repo_host) = @_;
	
	if ( defined($repo_host) )
	{
		logger("info> only refreshing repository host: $repo_host");
	}
	else
	{
		logger("info> refreshing datastructures on all repository hosts... (can take several minutes)");
	}

	#
	# delete tree items
	#
	$w_tree->delete('all');
	
	#
	# delete the info box from text posision 1.0 (start) to end - yes, the word!
	#
	$wtxt_zfsinfo->delete('1.0','end');
	
	#
	# fetch data and make the tree
	#
	zfs_repo_refresh($repo_host,'refresh');
	zfs_tree_make();
	$cfg{'tree_item'} = undef;
	zfs_tree_click();
	
	logger("info> refreshing datastructures completed.");

	$wcmd_clone_create->configure(-state => 'disabled');
	$wcmd_clone_delete->configure(-state => 'disabled');
}

#-------------------------------------------------------------------------------
# create clone from a snapshot
#-------------------------------------------------------------------------------
sub cmd_clone_create
{
	my $tree_item = $cfg{'tree_item'};
	
	my $properties = $cfg{'tree'}{$tree_item}{'properties'};
	my $repo_host = $$properties{'_tkzfs_host'};
	my $zpool = $$properties{'_tkzfs_zpool'};
	my $zvol = $$properties{'_tkzfs_zfs_name'}.'@'.$$properties{'_tkzfs_zfs_snapshot'};

	my $clone_name = "clone_$zvol";
	$clone_name =~ s/\@/_/g;
	my $itarget_enable = 1;
	my $itarget_initiator;
	my $itarget_portalgrp;
	
	my $wdlg = $w_main->DialogBox(
		-title => 'Create clone from snapshot',
		-buttons => ['Ok','Cancel'],
		-default_button => 'Ok',
		);
	$wdlg->add('Label',
		-text => 'Clone name:',
		)->grid(-row => 0, -column => 0);
	$wdlg->add('Entry',
		-width => 60,
		-textvariable => \$clone_name,
		)->grid(-row => 0, -column => 1);
		
	$wdlg->add('Label',
		-text => 'Enable iSCSI',
		)->grid(-row => 1, -column => 0);
	$wdlg->add('Checkbutton',
		-variable => \$itarget_enable,
		)->grid(-row => 1, -column => 1);
	
	my $this_host = hostname;
	my @initiators = zfs_repo_initiators($repo_host);
	foreach my $initiator ( @initiators )
	{
		if ( $initiator eq $cfg{'host'}{$this_host}{'zfsrepo_iscsitadm_initiator'} )
		{
			$itarget_initiator = $initiator;
		}
	}
	
	$wdlg->add('Label',
		-text => 'iSCSI Initiator',
		)->grid(-row => 2, -column => 0);
	$wdlg->add('BrowseEntry',
		-variable => \$itarget_initiator,
		-choices => \@initiators,
		)->grid(-row => 2, -column => 1);

	my @portalgrps = zfs_repo_portalgroups($repo_host);
	foreach my $portalgrp ( @portalgrps )
	{
		if ( $portalgrp eq $cfg{'repositories'}{$repo_host}{'iscsitadm_portalgrp'} )
		{
			$itarget_portalgrp = $portalgrp;
		}
	}

	$wdlg->add('Label',
		-text => 'iSCSI Portal Group',
		)->grid(-row => 3, -column => 0);
	$wdlg->add('BrowseEntry',
		-variable => \$itarget_portalgrp,
		-choices => \@portalgrps,
		)->grid(-row => 3, -column => 1);
	
	my $ans = $wdlg->Show();
	if ( $ans eq 'Ok' )
	{
		#
		# clone the snapshot
		#
		my $clone_check = "$repo_host/$zpool/$clone_name/";
		logger("clone> clone_check: $clone_check");
		if ( exists($cfg{'repositories'}{$repo_host}{'data'}{$clone_check}) )
		{
			logger("clone> duplicate found, refusing to create clone");
		}
		else
		{
			zfs_repo_clone_create($repo_host, $zpool, $zvol, $clone_name,
				$itarget_enable, $itarget_initiator, $itarget_portalgrp);
			cmd_refresh($repo_host);
		}
	}
}

#-------------------------------------------------------------------------------
# delete a cloned volume
#-------------------------------------------------------------------------------
sub cmd_clone_delete
{
	my $tree_item = $cfg{'tree_item'};
	
	my $properties = $cfg{'tree'}{$tree_item}{'properties'};
	my $repo_host = $$properties{'_tkzfs_host'};
	my $zpool = $$properties{'_tkzfs_zpool'};
	my $zvol = $$properties{'_tkzfs_zfs_name'};

	my $wdlg = $w_main->Dialog(
		-title => 'Confirmation',
		-text => "Delete the clone $cfg{'tree_item'}?\n\n".
			"The delete can take several minutes...",
		-default_button => 'No',
		-buttons => ['Yes','No'],
		);
	my $ans = $wdlg->Show();
	if ( $ans eq 'Yes' )
	{
		logger("clone> delete $cfg{'tree_item'}");
		
		zfs_repo_clone_delete($repo_host, $zpool, $zvol);
		cmd_refresh($repo_host);
	}
}

#-------------------------------------------------------------------------------
# map a cloned volume
#-------------------------------------------------------------------------------
sub cmd_map
{
	my $tree_item = $cfg{'tree_item'};

	if ( $tree_item eq "" )
	{
		logger("cmd_map_err> no item selected");
		return;
	}
	
	my $properties = $cfg{'tree'}{$tree_item}{'properties'};
	my $repo_host = $$properties{'_tkzfs_host'};

	#
	# dynamically dispatch the function call
	#
	my $this_host = hostname;
	my $iscsi_hba = $cfg{'host'}{$this_host}{'iscsi_hba'};
	logger("cmd_map> iscsi_hba: $iscsi_hba");
	$fn_dispatch{$iscsi_hba}{'map'}($w_main);

	#
	# refresh the gui
	#
	cmd_refresh($repo_host);
}

#-------------------------------------------------------------------------------
# unmap a cloned volume
#-------------------------------------------------------------------------------
sub cmd_umap
{
	my $tree_item = $cfg{'tree_item'};

	if ( $tree_item eq "" )
	{
		logger("cmd_map_err> no item selected");
		return;
	}

	#
	# dynamically dispatch the function call
	#
	my $this_host = hostname;
	my $iscsi_hba = $cfg{'host'}{$this_host}{'iscsi_hba'};
	logger("cmd_map> iscsi_hba: $iscsi_hba");
	$fn_dispatch{$iscsi_hba}{'umap'}($w_main);
}

