#
# tkzfs linux platform code
#
package tkzfslinux;
use strict;
require Exporter;

our @ISA = qw(Exporter);
our @EXPORT = qw(
);

use Data::Dumper;
use English;
use Sys::Hostname;

use FindBin;
use lib "$FindBin::Bin";
use config;
use tkzfsutil;
use tkzfsrepo;

#-------------------------------------------------------------------------------
# open-iscsi software initiator (works also with qla4xxx, man iscsiadm)
#-------------------------------------------------------------------------------
sub linux_lun_map
{
	my ($w_main) = @_;
	
	my $tree_item = $cfg{'tree_item'};
	
	my $properties = $cfg{'tree'}{$tree_item}{'properties'};
	my $repo_host = $$properties{'_tkzfs_host'};
	my $zpool = $$properties{'_tkzfs_zpool'};
	my $zvol = $$properties{'_tkzfs_zfs_name'};

	logger("map> using open_iscsi for zvol: $zvol");

	#
	# get the iqn name of the target from the repository
	#
	my $iqn_name = zfs_repo_iscsi_target_get_iname($repo_host, $zpool, $zvol);
	if ( !defined($iqn_name) || $iqn_name !~ /^iqn./ )
	{
		logger("map_lookup_err> unable to determine iscsi qualified name (iqn) from iscsi target, expecting iqn. name");
		return;
	}
	logger("map> using iqn: $iqn_name");


	#
	# lookup target in initiator and if not found add discovery record
	#
	logger("map> lookup target");
	if ( !open_iscsi_target_lookup($iqn_name) )
	{
		logger("map> target not found, adding discovery record");
		open_iscsi_discovery_add($cfg{'repositories'}{$repo_host}{'iscsi_target_ip'});
	}
	
	#
	# login to target
	#
	if ( !open_iscsi_session_lookup($iqn_name) )
	{
		logger("map> no active session found, login into target...");
		open_iscsi_target_login($iqn_name, $cfg{'repositories'}{$repo_host}{'iscsi_target_ip'});
	}
	logger("map> login to target complete");
	
	#
	# lookup GUID of LUN in the repository, one LUN per target
	#
	logger("map> lookup GUID of first LUN of selected target");
	my $guid = zfs_repo_iscsi_target_lun0_guid($repo_host, $zpool, $zvol);
	if ( !defined($guid) )
	{
		logger("map_err> unable to determine GUID from remote LUN");
		return;
	}
	logger("map> LUN GUID: $guid");

	#
	# find the device name and report that to the user
	#
	my $dev = linux_block_device_lookup($guid);
	logger("map> found device: $dev");
	if ( !defined($dev) )
	{
		logger("map_err> unable to find target device");
		return;
	}
	
	$w_main->messageBox(
		-message => "The target $iqn_name has an associated device of /dev/$dev",
		-title => "Device has been found...",
		-type => "Ok",
	);
}

#-------------------------------------------------------------------------------
# open-iscsi software initiator: lookup LUN
#-------------------------------------------------------------------------------
sub linux_lun_lookup
{
	my ($w_main) = @_;

	my $tree_item = $cfg{'tree_item'};
	
	my $properties = $cfg{'tree'}{$tree_item}{'properties'};
	my $repo_host = $$properties{'_tkzfs_host'};
	my $zpool = $$properties{'_tkzfs_zpool'};
	my $zvol = $$properties{'_tkzfs_zfs_name'};

	logger("lookup> using open_iscsi for zvol: $zvol");

	#
	# get the iqn name of the target from the repository
	#
	my $iqn_name = zfs_repo_iscsi_target_get_iname($repo_host, $zpool, $zvol);
	if ( !defined($iqn_name) || $iqn_name !~ /^iqn./ )
	{
		logger("lookup_err> unable to determine iscsi qualified name (iqn) from iscsi target, expecting iqn. name");
		return;
	}
	logger("lookup> using iqn: $iqn_name");
	
	#
	# look for an active session with this target's iqn
	#
	my $is_mapped = 0;
	if ( open_iscsi_session_lookup($iqn_name) )
	{
		$is_mapped = 1;
	}
	return $is_mapped;
}

#-------------------------------------------------------------------------------
# open-iscsi software initiator: unmap LUN
#-------------------------------------------------------------------------------
sub linux_lun_umap
{
	my ($w_main) = @_;
	
	my $tree_item = $cfg{'tree_item'};
	
	my $properties = $cfg{'tree'}{$tree_item}{'properties'};
	my $repo_host = $$properties{'_tkzfs_host'};
	my $zpool = $$properties{'_tkzfs_zpool'};
	my $zvol = $$properties{'_tkzfs_zfs_name'};

	logger("umap> using open_iscsi for zvol: $zvol");

	#
	# get the iqn name of the target from the repository
	#
	my $iqn_name = zfs_repo_iscsi_target_get_iname($repo_host, $zpool, $zvol);
	if ( !defined($iqn_name) || $iqn_name !~ /^iqn./ )
	{
		logger("umap_lookup_err> unable to determine iscsi qualified name (iqn) from iscsi target, expecting iqn. name");
		return;
	}

	#
	# look for an active session to logout from
	#
	if ( open_iscsi_session_lookup($iqn_name) )
	{
		logger("umap> found an active session to logout from...");
		open_iscsi_target_logout($iqn_name, $cfg{'repositories'}{$repo_host}{'iscsi_target_ip'});
	}
	
	logger("umap> complete");
}

#-------------------------------------------------------------------------------
# open-iscsi: target lookup
#-------------------------------------------------------------------------------
sub open_iscsi_target_lookup
{
	my ($iqn_name) = @_;
	
	my $cmd = "iscsiadm -m node";
	my $out = cmd_exec($cmd);
	my @lines = split(/\n/,$out);
	foreach my $line ( @lines )
	{
		if ( $line =~ /(iqn.*)/ )
		{
			my $iqn = $1;
			if ( $iqn eq $iqn_name )
			{
				return 1;
			}
		}
	}
	return 0;
}

#-------------------------------------------------------------------------------
# open-iscsi: discovery add
#-------------------------------------------------------------------------------
sub open_iscsi_discovery_add
{
	my ($portal_ip) = @_;

	my $cmd = "iscsiadm --mode discovery --type sendtargets --portal $portal_ip";
	cmd_exec($cmd);
}

#-------------------------------------------------------------------------------
# open-iscsi: target login
#-------------------------------------------------------------------------------
sub open_iscsi_target_login
{
	my($iqn_name, $portal_ip) = @_;
	
	my $cmd = "iscsiadm --mode node --targetname $iqn_name --portal $portal_ip --login";
	cmd_exec($cmd);
}

#-------------------------------------------------------------------------------
# open-iscsi: target logoout
#-------------------------------------------------------------------------------
sub open_iscsi_target_logout
{
	my($iqn_name, $portal_ip) = @_;
	
	my $cmd = "iscsiadm --mode node --targetname $iqn_name --portal $portal_ip --logout";
	cmd_exec($cmd);
}

#-------------------------------------------------------------------------------
# open-iscsi: session lookup
#-------------------------------------------------------------------------------
sub open_iscsi_session_lookup
{
	my ($iqn_name) = @_;
	
	my $cmd = "iscsiadm -m session";
	my $out = cmd_exec($cmd);
	my @lines = split(/\n/,$out);
	foreach my $line ( @lines )
	{
		if ( $line =~ /(iqn.*)/ )
		{
			my $iqn = $1;
			if ( $iqn eq $iqn_name )
			{
				return 1;
			}
		}
	}
	return 0;
}

#-------------------------------------------------------------------------------
# linux: lookup block device using sg_vpd --ident
#-------------------------------------------------------------------------------
sub linux_block_device_lookup
{
	my($guid) = @_;
	
	foreach my $blockdev (glob "/sys/block/sd*")
	{
		$blockdev =~ /(sd.*)$/;
		
		my $dev = $1;
		logger("block> $blockdev $dev 2> /dev/null");

		my $cmd = "sg_vpd --ident --quiet /dev/$dev";
		my $out = cmd_exec($cmd);
		my @lines = split(/\n/,$out);
		foreach my $line ( @lines )
		{
			if ( $line eq "0x$guid" )
			{
				return $dev;
			}
		}
	}
	return undef;
}

return 1;
