#
# tkzfs repository access code
#
package tkzfsrepo;
use strict;
require Exporter;

our @ISA = qw(Exporter);
our @EXPORT = qw(
 		zfs_repo_cmd
 		zfs_repo_refresh
 		zfs_repo_initiators
 		zfs_repo_portalgroups
 		zfs_repo_clone_create
 		zfs_repo_clone_delete
 		zfs_repo_iscsi_target_get_iname
 		zfs_repo_iscsi_target_lun0_guid
 		zfs_repo_version
	);

use Data::Dumper;
use English;

use FindBin;
use lib "$FindBin::Bin";
use config;
use tkzfsutil;

#-------------------------------------------------------------------------------
# given the current OS, determine the proper ssh command to reach a repository
#-------------------------------------------------------------------------------
sub zfs_repo_cmd
{
	my ($repo_host) = @_;
	
	return $cfg{'os'}{$OSNAME}{'ssh'}."$repo_host ~/santools/tkzfs/server";
}

#-------------------------------------------------------------------------------
# fetch a hash indexed by zfs volume/snap name, containing another
# hash of properties and property sources the idx is a list
# of zfs devices in proper order as listed by the zfs list command
#-------------------------------------------------------------------------------
sub zfs_repo_refresh
{
	my ($refresh_repo_host, $arg) = @_;

	#
	# clear out the tree structure and its pointers...
	#
	$cfg{'tree'} = undef;

	#
	# enumerate over the repositories that are enabled
	#
	foreach my $repo_name (keys %{$cfg{'repositories'}})
	{
		logger("repo> host: $repo_name (getting zfs properties)");
		my $repo = $cfg{'repositories'}{$repo_name};

		if ( ! $$repo{'enabled'} )
		{
			logger("repo> not enabled, skipping it");
			next;
		}

		if ( defined($refresh_repo_host) && $refresh_repo_host ne $repo_name )
		{
			logger("repo> not refreshing at this time");
			next;
		}
		
		my $cmd_repo = zfs_repo_cmd($repo_name);
		my $cmd = "$cmd_repo/zfs_tree.sh $arg";
		logger("repo> $cmd");
		
		#
		# the index contains the tree list items and its associated repo name
		# and the data hash contains the individual properties of each zfs item
		#
		$$repo{'index'} = undef;
		$$repo{'data'} = undef;
		
		open(FH,"$cmd|") || die "Unable to open $cmd: $!";
		while(<FH>)
		{
			my $line = $_;
			
			#
			# the line consists of name property value and property source
			#
			if ( $line =~ /^(.*)\t(.*)\t(.*)\t(.*)$/ )
			{
				my $name = $1;
				my $property = $2;
				my $value = $3;
				my $property_source =$4;
			
				#
				# blank values are caused by private properties like
				# iscsioptions, so hide them if they are blank
				#
				if ( $value eq '-' )
				{
					next;
				}

				#
				# the name consits of host/pool/zfsobject@snapshot [snapshot is optional]
				#
				if ( $name =~ /^(.*?)\/(.*?)\/(.*?)$/ )
				{
					my $host = $1;
					my $zpool = $2;
					my $zfs_name = $3;
					
					my $zfs_snapshot;
					if ( $zfs_name =~ /^(.*)@(.*)$/ )
					{
						$zfs_name = $1;
						$zfs_snapshot = $2;
					}
					
					if ( $host ne $repo_name )
					{
						logger("repo> reject record ($line) expecting host $repo_name");
						next;
					}
					
					#logger("host: $host zpool: $zpool zfs_name: $zfs_name zfs_snap: $zfs_snapshot");
					my $zfs_object = "$host/$zpool/$zfs_name/$zfs_snapshot";
					#logger("zfs_object: $zfs_object");
					
					$$repo{'data'}{$zfs_object}{'properties'}{'_tkzfs_host'} = $host;
					$$repo{'data'}{$zfs_object}{'properties'}{'_tkzfs_zpool'} = $zpool;
					$$repo{'data'}{$zfs_object}{'properties'}{'_tkzfs_zfs_name'} = $zfs_name;
					if ( defined($zfs_snapshot) )
					{
						$$repo{'data'}{$zfs_object}{'properties'}{'_tkzfs_zfs_snapshot'} = $zfs_snapshot;
					}

					$$repo{'data'}{$zfs_object}{'properties'}{$property} = $value;
					$$repo{'data'}{$zfs_object}{'prop_source'}{$property} = $property_source;

				}
			}
		}

		logger("repo> host: $repo_name loaded ".scalar(keys %{$$repo{'data'}})." objects.");
		
		#print Dumper \%cfg;
	}

	#
	# now build the tree data structure by which the repositories and their contents are presented
	#
	foreach my $repo_name (keys %{$cfg{'repositories'}})
	{
		#logger("repo> host: $repo_name (building tree structure)");
		my $repo = $cfg{'repositories'}{$repo_name};

		if ( ! defined($$repo{'data'}) )
		{
			next;
		}
		
		#
		# tree key (laid out to aid with sorting)
		#
		# g-elogic0_lv-srvrnt68-data-d_bk/zpool/host
		# g-elogic0_lv-srvrnt68-data-d_bk/zpool/host@2008-12-21_07:51:53
		#
		# tree displayname:
		#
		# g-elogic0_lv-srvrnt68-data-d_bk (zpool/host)
		# 	2008-12-21_07:51:53
		#
		
		foreach my $zfs_object ( keys %{$$repo{'data'}} )
		{
			my $zfs_properties = $$repo{'data'}{$zfs_object}{'properties'};
			
			my $zfs_name = $$zfs_properties{'_tkzfs_zfs_name'};
			my $zpool = $$zfs_properties{'_tkzfs_zpool'};
			my $host = $$zfs_properties{'_tkzfs_host'};

			my $tree_parent = "$zfs_name/$zpool/$host";
			my $tree_key = $tree_parent;
			my $tree_display = "$zfs_name ($zpool/$host)";
			if ( defined($$zfs_properties{'_tkzfs_zfs_snapshot'}) )
			{
				my $zfs_snapshot = $$zfs_properties{'_tkzfs_zfs_snapshot'};
				$tree_key = "$tree_parent\@$zfs_snapshot";
				$tree_display = "$zfs_snapshot";
			}
			
			#logger("repo> tree_key: $tree_key");
			
			#
			# establish tree key and display value and lay a pointer to the
			# actual properties hash of this tree item
			#
			$cfg{'tree'}{$tree_key}{'display'} = $tree_display;
			$cfg{'tree'}{$tree_key}{'parent'} = $tree_parent;
			$cfg{'tree'}{$tree_key}{'properties'} = $zfs_properties;
		}
	}
	#print Dumper \$cfg{'tree'};
	#exit;
}

#-------------------------------------------------------------------------------
# given a repository hostname, determine the registered initiators (it is assumed
# that each reposistory node is consistently configured).
#-------------------------------------------------------------------------------
sub zfs_repo_initiators
{
	my ($repo_host) = @_;
	
	my $cmd_repo = zfs_repo_cmd($repo_host);
	my $cmd = "$cmd_repo/iscsi_target_list_initiators.sh";
	logger("repo> $cmd");

	my $out = cmd_exec($cmd);
	my @initiators = split(/\n/,$out);

	return @initiators;
}

#-------------------------------------------------------------------------------
# given a repository hostname, determine the iscsi target portal groups
#-------------------------------------------------------------------------------
sub zfs_repo_portalgroups
{
	my ($repo_host) = @_;
	
	my $cmd_repo = zfs_repo_cmd($repo_host);
	my $cmd = "$cmd_repo/iscsi_target_list_portalgroups.sh";
	logger("repo> $cmd");

	my $out = cmd_exec($cmd);
	my @portalgrps = split(/\n/,$out);

	return @portalgrps;
}

#-------------------------------------------------------------------------------
# create a clone of a snapshot
#-------------------------------------------------------------------------------
sub zfs_repo_clone_create
{
	my ($repo_host, $zpool, $snapshot_name, $clone_name,
		$itarget_enable, $itarget_initiator, $itarget_portalgrp) = @_;
	
	my $cmd_repo = zfs_repo_cmd($repo_host);
	my $cmd = "$cmd_repo/zfs.sh clone $zpool/$snapshot_name $zpool/$clone_name";
	cmd_exec($cmd);
	
	if ( $itarget_enable )
	{
		$cmd = "$cmd_repo/zfs.sh set shareiscsi=on $zpool/$clone_name";
		cmd_exec($cmd);

		if ( $itarget_initiator ne "" )
		{
			$cmd = "$cmd_repo/iscsitadm.sh modify target --acl $itarget_initiator $zpool/$clone_name";
			cmd_exec($cmd);
		}
		if ( $itarget_portalgrp ne "" )
		{
			$cmd = "$cmd_repo/iscsitadm.sh modify target --tpgt $itarget_portalgrp $zpool/$clone_name";
			cmd_exec($cmd);
		}
	}
}

#-------------------------------------------------------------------------------
# delete a clone
#-------------------------------------------------------------------------------
sub zfs_repo_clone_delete
{
	my ($repo_host, $zpool, $zvol) = @_;
	
	my $cmd_repo = zfs_repo_cmd($repo_host);
	my $cmd = "$cmd_repo/zfs.sh destroy $zpool/$zvol";
	cmd_exec($cmd);
}

#-------------------------------------------------------------------------------
# get iscsitarget iqn name of a volume (zfs establishes a single taget per volume)
#-------------------------------------------------------------------------------
sub zfs_repo_iscsi_target_get_iname
{
	my ($repo_host, $zpool, $zvol) = @_;
	
	my $cmd_repo = zfs_repo_cmd($repo_host);
	my $cmd = "$cmd_repo/iscsi_target_get_iname.sh $zpool/$zvol";
	my $iname = cmd_exec($cmd);
	
	return $iname;
}

#-------------------------------------------------------------------------------
# get iscsitarget iqn name of a volume (zfs establishes a single taget per volume)
#-------------------------------------------------------------------------------
sub zfs_repo_iscsi_target_lun0_guid
{
	my ($repo_host, $zpool, $zvol) = @_;
	
	my $cmd_repo = zfs_repo_cmd($repo_host);
	my $cmd = "$cmd_repo/iscsi_target_lun0_guid.sh $zpool/$zvol";
	my $guid = cmd_exec($cmd);

	return $guid;
}

#-------------------------------------------------------------------------------
# version info
#-------------------------------------------------------------------------------
sub zfs_repo_version
{
	my ($repo_host) = @_;
	
	my $cmd_repo = zfs_repo_cmd($repo_host);
	my $cmd = "$cmd_repo/tkzfs_version.sh";
	my $repo_version = cmd_exec($cmd);

	return $repo_version;
}
