#!/usr/bin/perl

# ZFS Copy by Steffen Plotner, 2009, 2011

use strict;
use Data::Dumper;
use Getopt::Std;
use File::Copy;
use POSIX ":sys_wait_h";
use English;
use IO::Select;

use FindBin;
use lib "$FindBin::Bin/lib";
use lib "$FindBin::Bin/etc";

use common;
use config;

$cfg{'verbose'} = 1;

#
# handle command line options
#
my %opts;
getopts('hi:I:s:Ft:p',\%opts);
if ( defined($opts{'h'}) )
{
	print <<EOH;
$0 [-h] [-iI snap] -s [host]:zpool/zvolA\@snap [-F] -t [host]:zpool/zvolB
$0 [-h]            -s [host]:zpool/zvolA\@snap [-F] -t [host]:zpool -p

-h	help

-i	incremental data transfer from -i snap to -s snap (man zfs send)
-I	incremental data transfer from -I snap to -s snap transferring all
	intermediate snapshots as individual snapshots  (man zfs send)
-s	host name or IP address, source zpool and zfs object, such as zvol
-F	force rollback of last snapshot (man zfs receive)
-t	host name or IP address, target zpool and target object name
-p	preserve all zfs properties (implies zfs send -R)  (man zfs send,receive)

The source object must have at least one snapshot.

If you use -p, then internally we use zfs send -R (includes all snapshots up to
zvolA\@snap) and zfs receive -d which implies that the source and destination
pool cannot be the same, on different hosts they can, but not on the same host.

Without the -p the data is copied, the properties are not copied and the above
restriction does not apply, i.e. you can copy a zvol on the same hosts within
the same zpool.

If you want to copy starting with a specific snapshot and then incremental
snapshots in the future from a source to destination use two commands:

	1) zfs_copy.plx            -s [host]:zpool/zvolA\@snap -t [host]:zpool/zvolB
	2) zfs_copy.plx [-iI snap] -s [host]:zpool/zvolA\@snap -t [host]:zpool/zvolB

The first command creates a target zvol and the snapshot name from the
source(!). Then you can send incremental snapshots. This implies that -p
cannot be used here, because -p would include ALL prior snapshots.
EOH
	exit(0);
}
logger("copy> zfs object copy");

my $incremental_snap_type;
my $incremental_snap_name;
if ( defined($opts{'i'}) )
{
	$incremental_snap_type = "-i";
	$incremental_snap_name = $opts{'i'};
}
if ( defined($opts{'I'}) )
{
	$incremental_snap_type = "-I";
	$incremental_snap_name = $opts{'I'};
}
if ( defined($incremental_snap_type) )
{
	logger("copy> incremental snap type: $incremental_snap_type");
	logger("copy> incremental snap name: $incremental_snap_name");
}

my $force_target;
if ( defined($opts{'F'}) )
{
	$force_target = "-F";
	logger("copy> force target rollback: $force_target");
}

my $preserve_properties = 0;
if ( defined($opts{'p'}) )
{
	$preserve_properties = 1;
}
if ( defined($preserve_properties) )
{
	logger("copy> preserve properties: $preserve_properties");
}

my $recent_data = 0;
if ( defined($opts{'d'}) )
{
	$recent_data = 1;
}
logger("copy> recent data: $recent_data");

my $source_parm;
if ( defined($opts{'s'}) )
{
	$source_parm = $opts{'s'};
}

my $target_parm;
if ( defined($opts{'t'}) )
{
	$target_parm = $opts{'t'};
}

my %source_session = parse_parm($source_parm);
my %target_session = parse_parm($target_parm);
if ( $source_session{'zpool'} eq "" && $target_session{'zpool'} eq "" )
{
	logger_err("copy_err> use -h");
	exit(1);
}
logger("copy> source host: $source_session{'host'} ".
	"zpool: $source_session{'zpool'} zvol: $source_session{'zvol'}");
logger("copy> target host: $target_session{'host'} ".
	"zpool: $target_session{'zpool'} zvol: $target_session{'zvol'}");

my $cpid;
$SIG{INT} = \&sig_int;

#
# zfs receive options
#
my $zfs_receive_cmd;
my $zfs_receive_options = "-v $force_target $target_session{'zpool'}/$target_session{'zvol'}";
if ( $preserve_properties )
{
	$zfs_receive_options = "-v $force_target -d $target_session{'zpool'}";
}
$zfs_receive_cmd = "zfs receive $zfs_receive_options";
$zfs_receive_cmd =~ s/\:/\\\:/g;

#
# if target has a host involved, then setup that communication endpoint first
#
my $tcp_listen_port;
if ( defined($target_session{'host'}) )
{
	#
	# on the remote host, ensure that the tcp stack is optimized
	#
	my $cmd = "ssh $target_session{'host'} $cfg{'zfscopy'}{'init_channel'}";
	logger("copy> cmd=$cmd");
	system($cmd);
	
	#
	# on the remote host, get a unique new port number to listen on
	#
	$cmd = "ssh $target_session{'host'} $cfg{'zfscopy'}{'port_inc'}";
	logger("copy> cmd=$cmd");
	$tcp_listen_port = `$cmd`;
	chomp($tcp_listen_port);
	logger("copy> tcp_listen_port=$tcp_listen_port");
	if ( $tcp_listen_port eq "" )
	{
		logger_err("copy_err> unable to obtain tcp listen port");
		exit(1);
	}

	#
	# one-way pipe from child to parent (for stdout and stderr)
	#
	pipe(FROM_CHILD,TO_PARENT);

	#
	# fork the child
	#
	$cpid = fork();
	if ( !defined($cpid) )
	{
		logger_err("copy_err> unable to fork");
		exit 1;
	}
	if ( $cpid == 0 )
	{
		#
		# child process, intent is to invoke the tcp listener
		#
		logger("child> forked copy process pid: $$");
		
		#
		# merge stdout and err, unlikely that they will intermix
		# with the commands below, otherwise we need 2 pipes and
		# a select in the parent...
		#
		open(STDOUT,">&TO_PARENT") || die ("open: $!");
		open(STDERR,">&STDOUT") || die ("open: $!");

		#
		# close the other half of the pipe
		#
		close(FROM_CHILD);

		#
		# hot pipes
		#
		select STDERR; $OUTPUT_AUTOFLUSH = 1;
		select STDOUT; $OUTPUT_AUTOFLUSH = 1;

		#
		# this complex (sorry) thing does a ssh session invokes socat
		# and pipes that into zfs receive. 
		#
		$cmd = "ssh $target_session{'host'} ".
			"'".
			"$cfg{'zfscopy'}{'socat'} -u -b $cfg{'zfscopy'}{'socat_buf'} ".
			"TCP4-LISTEN:$tcp_listen_port,$cfg{'zfscopy'}{'socat_rcv'} ".
			"SYSTEM:\"$zfs_receive_cmd\"".
			"'";
		print "cmd=$cmd\n";
		exec($cmd);
	}
	
	#
	# back at the parent, close the other half of the pipe
	#
	close(TO_PARENT);
	
	#
	# just wait for the child to open the socket
	#
	logger("copy> parent waiting for child (cpid=$cpid) to settle...");
	sleep 3;

	logger("copy> we are ready using port number: $tcp_listen_port");
	child_readline_dump();	
}

#
# on current system (this could be in reference to source_session local or
# target_session local), ensure that the tcp stack is optimized
#
my $cmd = "$cfg{'zfscopy'}{'init_channel'}";
if ( defined($source_session{'host'}) )
{
	$cmd = "ssh $source_session{'host'} $cmd";
}
logger("copy> cmd=$cmd");
system($cmd);

#
# source ssh prefix (requires single quote to preserve other quotes)
#
my $ssh;
if ( defined($source_session{'host'}) )
{
	$ssh = "ssh $source_session{'host'} '";
}

#
# source zfs send (optional -i/-I followed by source pool and source zvol+snap)
#
my $zfs_send_cmd;
my $zfs_send_options = "$incremental_snap_type $incremental_snap_name ".
	"$source_session{'zpool'}/$source_session{'zvol'}";
if ( $preserve_properties )
{
	$zfs_send_options = "-R $source_session{'zpool'}/$source_session{'zvol'}";
}
$zfs_send_cmd = "zfs send $zfs_send_options";
$zfs_send_cmd =~ s/\:/\\\:/g;

#
# source command
#
if ( defined($source_session{'host'}) || defined($target_session{'host'}) )
{
	$cmd = "$ssh ".
		"$cfg{'zfscopy'}{'socat'} -u -b $cfg{'zfscopy'}{'socat_buf'} ".
		"SYSTEM:\"$zfs_send_cmd\" ".
		"TCP4:$target_session{'host'}:$tcp_listen_port,$cfg{'zfscopy'}{'socat_snd'}";
}
else
{
	$cmd = "$zfs_send_cmd | $zfs_receive_cmd";
}

#
# source ssh suffix
#
if ( defined($source_session{'host'}) )
{
	$cmd .= "'";
}
logger("copy> cmd=$cmd");

my $ccode = system($cmd);
my $parent_exitcode = WEXITSTATUS($ccode);

#
# now that the parent is done, fetch the child's output
#
if ( defined($target_session{'host'}) )
{
	#
	# dump what the child has for us in stdout
	#
	child_readline_dump();
	
	#
	# wait for child to quit
	#
	logger("copy> waiting for child to exit...");
	wait;
	my $cpid_exitcode = WEXITSTATUS($?);
	
	#
	# and dump the rest of the child's stdout
	#
	child_readline_dump();

	logger("copy> parent exit code: $parent_exitcode");
	logger("copy> child exit code: $cpid_exitcode");

	if ( $parent_exitcode != 0 || $cpid_exitcode != 0 )
	{
		logger_err("copy_err> parent exit code: $parent_exitcode, child exit code: $cpid_exitcode");
		exit(1);
	}
}

logger("copy> completed");

exit(0);



#-------------------------------------------------------------------------------
# on break, clean up
#-------------------------------------------------------------------------------
sub sig_int
{
	logger("sig_int> cleanup");
	kill TERM => $cpid;
	logger("sig_int> complete");
	exit(0);
}

#-------------------------------------------------------------------------------
# read from child process stdout and present it
#-------------------------------------------------------------------------------
sub child_readline_dump
{
	my $line;
	while ( ($line = child_readline()) )
	{
		chomp($line);
		logger("child_out> $line");
	}
}

#-------------------------------------------------------------------------------
# read from child process stdout (be aware of blocking)
#-------------------------------------------------------------------------------
sub child_readline
{
	my $sel = IO::Select->new();
	$sel->add(\*FROM_CHILD);
	
	my $line;
	if ( $sel->can_read(.5) )
	{
		$line = <FROM_CHILD>;
	}
	return $line;
}

#-------------------------------------------------------------------------------
# given a [host]:zpool/zvol object, return the individual pieces
#-------------------------------------------------------------------------------
sub parse_parm
{
	my($parm) = @_;
	
	my %session;
	
	if ( $parm =~ /^(.*?):(.*)\/(.*)$/ )
	{
		$session{'host'} = $1;
		$session{'zpool'} = $2;
		$session{'zvol'} = $3;
	}
	elsif ( $parm =~ /^(.*)\/(.*)$/ )
	{
		$session{'zpool'} = $1;
		$session{'zvol'} = $2;
	}
	elsif ( $parm =~ /^(.*?):(.*)$/ )
	{
		$session{'host'} = $1;
		$session{'zpool'} = $2;
	}
	elsif ( $parm =~ /^(.*)$/ )
	{
		$session{'zpool'} = $1;
	}
	else
	{
		logger_err("copy> bad parameter: $parm (".
			"should be optional host: followed by pool/object)");
		exit(1);
	}
	return %session;
}


