#!/usr/bin/perl

# ZFS Report by Steffen Plotner, 2011

#
# was used by the SAN Report Drawing in Visio
# has been modified to support zvol size reporting
# and zpool size reporting for systemd.
#

use strict;
use Data::Dumper;
use Getopt::Std;

use FindBin;
use lib "$FindBin::Bin/lib";
use lib "$FindBin::Bin/etc";

use common;
use config;
use SunZfs;

#
# handle command line options
#
my %opts;
getopts('hvr',\%opts);
if ( defined($opts{'h'}) )
{
	print <<EOH;
$0 [-h] [-v]

-h	help
-v	verbose
-r	human readable report
EOH
	exit(0);
}

if ( defined($opts{'v'}) )
{
	$cfg{'verbose'} = $opts{'v'};
}

my $is_report = 0;
if ( defined($opts{'r'}) )
{
	$is_report = 1;
}

my $host = 'localhost';
logger("report> running on host: $host");
my $gb = 1024 * 1024 * 1024;

#
# zvol data
#
my %data;

my @zvols = zfs_list_zvols($host);
foreach my $zvol (@zvols)
{
	my $props = zfs_props($host, $zvol, 1);
	
	$data{$zvol}{'volsize'} = $$props{'volsize'};
	$data{$zvol}{'compressratio'} = $$props{'compressratio'};
	$data{$zvol}{'used'} = $$props{'used'};

	my @snaps = zfs_snapshots($host, $zvol);
	
	$data{$zvol}{'snapcount'} = scalar(@snaps);
}

#
# zvol report
#
my @zvol_report;

my $total_zvol_volsize;
my $total_zvol_used;
my $total_zvol_snapcount;

my $zvol_line_format = "%-60s %19s %8s  %5s %8s %5s";

foreach my $zvol (keys %data)
{
	my ($pool, $vol) = split('/', $zvol);

	my $volsize = $data{$zvol}{'volsize'};
	my $compress = $data{$zvol}{'compressratio'};
	my $used = $data{$zvol}{'used'};
	my $snapcount = $data{$zvol}{'snapcount'};
	
	my $line = sprintf($zvol_line_format,
		$vol,
		$pool, 
		sprintf("%5d GB",$volsize/$gb),
		$compress,
		sprintf("%5d GB",$used/$gb),
		sprintf("%5d",$snapcount),
		);
	push @zvol_report, $line;
	
	$total_zvol_volsize += $volsize;
	$total_zvol_used += $used;
	$total_zvol_snapcount += $snapcount;
	
	if ( !$is_report )
	{
		my $compress_number = $compress;
		$compress_number =~ s/x//g;
		print "zvol> $pool-$vol $volsize $compress_number $used $snapcount\n";
	}
}

#
# zpool data/report
#
my @zpool_report;
my $zpools = zpools_stats();

my $total_zpool_size;
my $total_zpool_used;

my $zpool_line_format = "%-19s %8s %8s";
foreach my $zpool (keys %{$zpools})
{
	my $pool_size = $$zpools{$zpool}{'size'};
	my $pool_used = $$zpools{$zpool}{'used'};
	
	my $line = sprintf($zpool_line_format,
		$zpool,
		sprintf("%5d GB",$pool_size/$gb),
		sprintf("%5d GB",$pool_used/$gb),
		);
	push @zpool_report, $line;
	
	$total_zpool_size += $pool_size;
	$total_zpool_used += $pool_used;

	if ( !$is_report )
	{
		print "zpool> $zpool $pool_size $pool_used\n";
	}
}

if ( $is_report )
{
	printf("$zvol_line_format\n",
		"zvol",
		"zpool",
		"disk",
		"comp",
		"used",
		"snap",
		);
	printf("$zvol_line_format\n",'-'x60,'-'x19,'-'x8,'-'x5,'-'x8,'-'x5);
	print join "\n", sort @zvol_report;
	print "\n";
	printf("$zvol_line_format\n",'-'x60,'-'x19,'-'x8,'-'x5,'-'x8,'-'x5);

	printf($zvol_line_format,
		"zvol grant total", 
		"", 
		sprintf("%5d GB",$total_zvol_volsize/$gb),
		"",
		sprintf("%5d GB",$total_zvol_used/$gb),
		sprintf("%5d",$total_zvol_snapcount),
		);
	print "\n\n";
	
	printf("$zpool_line_format\n",
		"zpool",
		"size",
		"used",
		);
	printf("$zpool_line_format\n",'-'x19,'-'x8,'-'x8);
	print join "\n", sort @zpool_report;
	print "\n";
	printf("$zpool_line_format\n",'-'x19,'-'x8,'-'x8);
	printf($zpool_line_format,
		"zpool grant total", 
		sprintf("%5d GB",$total_zpool_size/$gb),
		sprintf("%5d GB",$total_zpool_used/$gb),
		);
	print "\n\n";
}

exit(0);


#-------------------------------------------------------------------------------
# gather all zpool stats
#-------------------------------------------------------------------------------
sub zpools_stats
{
	my $cmd = "zpool list -H -o name,size,used";
	logger("report> cmd: $cmd");
	
	my %zpools;
	my @out = `$cmd`;
	foreach my $line (@out)
	{
		if ($line =~ /(\S+)\t(\S+)\t(\S+)/ )
		{
			my $pool = $1;
			my $size = $2;
			my $used = $3;
			$zpools{$pool}{'size'} = human_to_computer_bytes($size);
			$zpools{$pool}{'used'} = human_to_computer_bytes($used);
		}
	}
	return(\%zpools);
}

#-------------------------------------------------------------------------------
# the zpool command only provides human readable storage sizes
# convert them to bytes
#-------------------------------------------------------------------------------
sub human_to_computer_bytes
{
	my($n) = @_;
	
	my $factor_char;
	my $bytes;
	if ( $n =~ /^(.*?)(.)$/ )
	{
		my $num = $1;
		my $factor_char = $2;
		
		if ( $factor_char eq 'K' )
		{
			$num *= 1024;
		}
		elsif ( $factor_char eq 'M' )
		{
			$num *= 1024 * 1024;
		}
		elsif ( $factor_char eq 'G' )
		{
			$num *= 1024 * 1024 * 1024;
		}
		elsif ( $factor_char eq 'T' )
		{
			$num *= 1024 * 1024 * 1024 * 1024;
		}
		else
		{
			printf STDERR "unknown unit?? $n: num: $num factor_char=$factor_char\n";
		}
		$bytes = int($num);
		
		#print "$n: $1, $2 $bytes\n";
	}
	return $bytes;
}

exit 0;

